# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import re
import sys
import time
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from graphlib import CycleError, TopologicalSorter
from pathlib import Path
from time import sleep
from typing import Any, cast, final

from cognite.client.data_classes import (
    filters,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelList,
    Edge,
    EdgeApply,
    EdgeApplyList,
    EdgeApplyResultList,
    EdgeList,
    Node,
    NodeApply,
    NodeApplyList,
    NodeApplyResultList,
    NodeList,
    Space,
    SpaceApply,
    SpaceApplyList,
    SpaceList,
    View,
    ViewApply,
    ViewApplyList,
    ViewList,
)
from cognite.client.data_classes.data_modeling.graphql import DMLApplyResult
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    DataModelId,
    EdgeId,
    NodeId,
    ViewId,
)
from cognite.client.data_classes.data_modeling.views import ReverseDirectRelationApply
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console
from rich.markup import escape
from rich.panel import Panel

from cognite_toolkit._cdf_tk._parameters import ANY_INT, ANY_STR, ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.graphql_data_models import (
    GraphQLDataModel,
    GraphQLDataModelList,
    GraphQLDataModelWrite,
    GraphQLDataModelWriteList,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING, HAS_DATA_FILTER_LIMIT
from cognite_toolkit._cdf_tk.exceptions import GraphQLParseError, ToolkitCycleError, ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.loaders._base_loaders import (
    ResourceContainerLoader,
    ResourceLoader,
)
from cognite_toolkit._cdf_tk.resource_classes import SpaceYAML
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    GraphQLParser,
    calculate_hash,
    in_dict,
    load_yaml_inject_variables,
    quote_int_value_by_key_in_yaml,
    safe_read,
    to_diff,
    to_directory_compatible,
)
from cognite_toolkit._cdf_tk.utils.cdf import iterate_instances
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_identifiable, dm_identifier
from cognite_toolkit._cdf_tk.utils.tarjan import tarjan

from .auth_loaders import GroupAllScopedLoader


@final
class SpaceLoader(ResourceContainerLoader[str, SpaceApply, Space, SpaceApplyList, SpaceList]):
    item_name = "nodes and edges"
    folder_name = "data_models"
    filename_pattern = r"^.*space$"
    resource_cls = Space
    resource_write_cls = SpaceApply
    list_write_cls = SpaceApplyList
    list_cls = SpaceList
    kind = "Space"
    yaml_cls = SpaceYAML
    dependencies = frozenset({GroupAllScopedLoader})
    _doc_url = "Spaces/operation/ApplySpaces"
    delete_recreate_limit_seconds: int = 10

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None) -> None:
        super().__init__(client, build_dir, console)
        self._deleted_time_by_id: dict[str, float] = {}

    @property
    def display_name(self) -> str:
        return "spaces"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SpaceApply] | None, read_only: bool
    ) -> list[Capability] | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        return [DataModelsAcl(actions, DataModelsAcl.Scope.All())]

    @classmethod
    def get_id(cls, item: SpaceApply | Space | dict) -> str:
        if isinstance(item, dict):
            return item["space"]
        return item.space

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"space": id}

    def create(self, items: Sequence[SpaceApply]) -> SpaceList:
        for item in items:
            item_id = self.get_id(item)
            if item_id in self._deleted_time_by_id:
                elapsed_since_delete = time.perf_counter() - self._deleted_time_by_id[item_id]
                if elapsed_since_delete < self.delete_recreate_limit_seconds:
                    time.sleep(self.delete_recreate_limit_seconds - elapsed_since_delete)
        return self.client.data_modeling.spaces.apply(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SpaceList:
        return self.client.data_modeling.spaces.retrieve(ids)

    def update(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        existing = self.client.data_modeling.spaces.retrieve(ids)
        is_global = {space.space for space in existing if space.is_global}
        if is_global:
            print(
                f"  [bold yellow]WARNING:[/] Spaces {list(is_global)} are global and cannot be deleted, skipping delete, for these."
            )
        to_delete = [space for space in ids if space not in is_global]
        deleted = self.client.data_modeling.spaces.delete(to_delete)
        for item_id in to_delete:
            self._deleted_time_by_id[item_id] = time.perf_counter()
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Space]:
        if space:
            return self.client.data_modeling.spaces.retrieve([space])
        else:
            return iter(self.client.data_modeling.spaces)

    def count(self, ids: SequenceNotStr[str]) -> int:
        # Bug in spec of aggregate requiring view_id to be passed in, so we cannot use it.
        # When this bug is fixed, it will be much faster to use aggregate.
        existing = self.client.data_modeling.spaces.retrieve(ids)

        return sum(len(batch) for batch in self._iterate_over_nodes(existing)) + sum(
            len(batch) for batch in self._iterate_over_edges(existing)
        )

    def drop_data(self, ids: SequenceNotStr[str]) -> int:
        existing = self.client.data_modeling.spaces.retrieve(ids)
        if not existing:
            return 0
        print(f"[bold]Deleting existing data in spaces {ids}...[/]")
        nr_of_deleted = 0
        for edge_ids in self._iterate_over_edges(existing):
            self.client.data_modeling.instances.delete(edges=edge_ids)
            nr_of_deleted += len(edge_ids)
        for node_ids in self._iterate_over_nodes(existing):
            self.client.data_modeling.instances.delete(nodes=node_ids)
            nr_of_deleted += len(node_ids)
        return nr_of_deleted

    def _iterate_over_nodes(self, spaces: SpaceList) -> Iterable[list[NodeId]]:
        is_space: filters.Filter
        if len(spaces) == 0:
            return
        elif len(spaces) == 1:
            is_space = filters.Equals(["node", "space"], spaces[0].as_id())
        else:
            is_space = filters.In(["node", "space"], spaces.as_ids())
        for instances in self.client.data_modeling.instances(
            chunk_size=1000, instance_type="node", filter=is_space, limit=-1
        ):
            yield instances.as_ids()

    def _iterate_over_edges(self, spaces: SpaceList) -> Iterable[list[EdgeId]]:
        is_space: filters.Filter
        if len(spaces) == 0:
            return
        elif len(spaces) == 1:
            is_space = filters.Equals(["edge", "space"], spaces[0].as_id())
        else:
            is_space = filters.In(["edge", "space"], spaces.as_ids())
        for instances in self.client.data_modeling.instances(
            chunk_size=1000, instance_type="edge", limit=-1, filter=is_space
        ):
            yield instances.as_ids()


class ContainerLoader(
    ResourceContainerLoader[ContainerId, ContainerApply, Container, ContainerApplyList, ContainerList]
):
    item_name = "nodes and edges"
    folder_name = "data_models"
    filename_pattern = r"^.*container$"
    resource_cls = Container
    resource_write_cls = ContainerApply
    list_cls = ContainerList
    list_write_cls = ContainerApplyList
    kind = "Container"
    dependencies = frozenset({SpaceLoader})
    _doc_url = "Containers/operation/ApplyContainers"

    @property
    def display_name(self) -> str:
        return "containers"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ContainerApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: ContainerApply | Container | dict) -> ContainerId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return ContainerId(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ContainerId) -> dict[str, Any]:
        return id.dump(include_type=False)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]
        # Note that we are very careful in the code below to not raise an exception if the
        # item is not properly formed. If that is the case, an appropriate warning will be given elsewhere.
        for prop in item.get("properties", {}).values():
            if not isinstance(prop, dict):
                continue
            prop_type = prop.get("type", {})
            if isinstance(prop_type, dict) and prop_type.get("type") == "direct":
                if isinstance(prop_type.get("container"), dict):
                    container = prop_type["container"]
                    if "space" in container and "externalId" in container and container.get("type") == "container":
                        yield (
                            ContainerLoader,
                            ContainerId(space=container["space"], external_id=container["externalId"]),
                        )

    def dump_resource(self, resource: Container, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        for key in ["constraints", "indexes"]:
            if not dumped.get(key) and key not in local:
                # Set to empty dict by server.
                dumped.pop(key, None)
        local_prop_by_id = local.get("properties", {})
        for prop_id, cdf_prop in dumped.get("properties", {}).items():
            if prop_id not in local_prop_by_id:
                continue
            local_prop = local_prop_by_id[prop_id]
            for key, default in [("immutable", False), ("autoIncrement", False), ("nullable", True)]:
                if cdf_prop.get(key) is default and key not in local_prop:
                    cdf_prop.pop(key, None)
            cdf_type = cdf_prop.get("type", {})
            local_type = local_prop.get("type", {})
            for key, type_default in [("list", False), ("collation", "ucs_basic")]:
                if cdf_type.get(key) == type_default and key not in local_type:
                    cdf_type.pop(key, None)
        if "usedFor" not in local:
            dumped.pop("usedFor", None)
        return dumped

    def create(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self.client.data_modeling.containers.apply(items)

    def retrieve(self, ids: SequenceNotStr[ContainerId]) -> ContainerList:
        return self.client.data_modeling.containers.retrieve(cast(Sequence, ids))

    def update(self, items: Sequence[ContainerApply]) -> ContainerList:
        updated = self.create(items)
        # The API might silently fail to update a container.
        updated_by_id = {item.as_id(): item for item in updated}
        for local in items:
            item_id = local.as_id()
            local_dict = local.dump()
            if item_id not in updated_by_id:
                raise CogniteAPIError(
                    f"The container {item_id} was not updated. You might need to delete and recreate it.",
                    code=500,
                )
            cdf_dict = self.dump_resource(updated_by_id[item_id], local_dict)
            if cdf_dict != local_dict:
                is_verbose = "-v" in sys.argv or "--verbose" in sys.argv
                if is_verbose:
                    print(
                        Panel(
                            "\n".join(to_diff(cdf_dict, local_dict)),
                            title=f"{self.display_name}: {item_id}",
                            expand=False,
                        )
                    )
                suffix = "" if is_verbose else " (use -v for more info)"
                HighSeverityWarning(
                    f"The container {item_id} was not updated. You might need to delete and recreate it{suffix}."
                ).print_warning()
        return updated

    def delete(self, ids: SequenceNotStr[ContainerId]) -> int:
        deleted = self.client.data_modeling.containers.delete(cast(Sequence, ids))
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Container]:
        return iter(self.client.data_modeling.containers(space=space))

    def count(self, ids: SequenceNotStr[ContainerId]) -> int:
        # Bug in spec of aggregate requiring view_id to be passed in, so we cannot use it.
        # When this bug is fixed, it will be much faster to use aggregate.
        existing_containers = self.client.data_modeling.containers.retrieve(cast(Sequence, ids))
        return sum(len(batch) for batch in self._iterate_over_nodes(existing_containers)) + sum(
            len(batch) for batch in self._iterate_over_edges(existing_containers)
        )

    def drop_data(self, ids: SequenceNotStr[ContainerId]) -> int:
        nr_of_deleted = 0
        existing_containers = self.client.data_modeling.containers.retrieve(cast(Sequence, ids))
        for node_ids in self._iterate_over_nodes(existing_containers):
            self.client.data_modeling.instances.delete(nodes=node_ids)
            nr_of_deleted += len(node_ids)
        for edge_ids in self._iterate_over_edges(existing_containers):
            self.client.data_modeling.instances.delete(edges=edge_ids)
            nr_of_deleted += len(edge_ids)
        return nr_of_deleted

    def _iterate_over_nodes(self, containers: ContainerList) -> Iterable[list[NodeId]]:
        container_ids = [container.as_id() for container in containers if container.used_for in ["node", "all"]]
        if not container_ids:
            return
        for container_id_chunk in self._chunker(container_ids, HAS_DATA_FILTER_LIMIT):
            is_container = filters.HasData(containers=container_id_chunk)
            for instances in self.client.data_modeling.instances(
                chunk_size=1000, instance_type="node", filter=is_container, limit=-1
            ):
                yield instances.as_ids()

    def _iterate_over_edges(self, containers: ContainerList) -> Iterable[list[EdgeId]]:
        container_ids = [container.as_id() for container in containers if container.used_for in ["edge", "all"]]
        if not container_ids:
            return

        for container_id_chunk in self._chunker(container_ids, HAS_DATA_FILTER_LIMIT):
            is_container = filters.HasData(containers=container_id_chunk)
            for instances in self.client.data_modeling.instances(
                chunk_size=1000, instance_type="edge", limit=-1, filter=is_container
            ):
                yield instances.as_ids()

    @staticmethod
    def _chunker(seq: Sequence, size: int) -> Iterable[Sequence]:
        return (seq[pos : pos + size] for pos in range(0, len(seq), size))

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        output = super().get_write_cls_parameter_spec()
        # In the SDK this is called isList, while in the API it is called list.
        output.discard(
            ParameterSpec(
                ("properties", ANY_STR, "type", "isList"), frozenset({"bool"}), is_required=True, _is_nullable=False
            )
        )
        output.add(
            ParameterSpec(
                ("properties", ANY_STR, "type", "list"), frozenset({"bool"}), is_required=True, _is_nullable=False
            )
        )
        # The parameters below are used by the SDK to load the correct class, and ase thus not part of the init
        # that the spec is created from, so we need to add them manually.
        output.update(
            ParameterSpecSet(
                {
                    ParameterSpec(
                        ("properties", ANY_STR, "type", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        # direct relations with constraint
                        ("properties", ANY_STR, "type", "container", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("constraints", ANY_STR, "constraintType"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("constraints", ANY_STR, "require", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("indexes", ANY_STR, "indexType"), frozenset({"str"}), is_required=True, _is_nullable=False
                    ),
                }
            )
        )
        return output

    @classmethod
    def as_str(cls, id: ContainerId) -> str:
        return to_directory_compatible(f"{id.space}_{id.external_id}")


class ViewLoader(ResourceLoader[ViewId, ViewApply, View, ViewApplyList, ViewList]):
    folder_name = "data_models"
    filename_pattern = r"^.*view$"
    resource_cls = View
    resource_write_cls = ViewApply
    list_cls = ViewList
    list_write_cls = ViewApplyList
    kind = "View"
    dependencies = frozenset({SpaceLoader, ContainerLoader})
    _doc_url = "Views/operation/ApplyViews"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None,
        topological_sort_implements: bool = False,
    ) -> None:
        super().__init__(client, build_dir, console)
        self._topological_sort_implements = topological_sort_implements
        self._view_by_id: dict[ViewId, View] = {}

    @property
    def display_name(self) -> str:
        return "views"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ViewApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: ViewApply | View | dict) -> ViewId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return ViewId(space=item["space"], external_id=item["externalId"], version=str(item["version"]))

        return ViewId(item.space, item.external_id, str(item.version))

    @classmethod
    def dump_id(cls, id: ViewId) -> dict[str, Any]:
        return id.dump(include_type=False)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]
        if isinstance(implements := item.get("implements", []), list):
            for parent in implements:
                if not isinstance(parent, dict):
                    continue
                if parent.get("type") == "view" and in_dict(["space", "externalId", "version"], parent):
                    yield (
                        ViewLoader,
                        ViewId(parent["space"], parent["externalId"], str(v) if (v := parent.get("version")) else None),
                    )
        for prop in item.get("properties", {}).values():
            if (container := prop.get("container", {})) and container.get("type") == "container":
                if in_dict(("space", "externalId"), container):
                    yield ContainerLoader, ContainerId(container["space"], container["externalId"])
            for key, dct_ in [("source", prop), ("edgeSource", prop), ("source", prop.get("through", {}))]:
                if source := dct_.get(key, {}):
                    if source.get("type") == "view" and in_dict(("space", "externalId", "version"), source):
                        yield (
                            ViewLoader,
                            ViewId(
                                source["space"], source["externalId"], str(v) if (v := source.get("version")) else None
                            ),
                        )
                    elif source.get("type") == "container" and in_dict(("space", "externalId"), source):
                        yield ContainerLoader, ContainerId(source["space"], source["externalId"])

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def dump_resource(self, resource: View, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if not dumped.get("properties") and not local.get("properties"):
            if "properties" in local:
                # In case the properties is an empty dict, we still want to keep it in the dump.
                # such that the dumped evaluates to the same as the local.
                dumped["properties"] = local["properties"]
            else:
                dumped.pop("properties", None)
        if not dumped.get("implements") and not local.get("implements"):
            if "implements" in local:
                # In case the implements is an empty list, we still want to keep it in the dump.
                # such that the dumped evaluates to the same as the local.
                dumped["implements"] = local["implements"]
            else:
                dumped.pop("implements", None)
        if resource.implements and len(resource.implements) > 1 and self._topological_sort_implements:
            # This is a special case that we want to do when we run the cdf dump datamodel command.
            # The issue is as follows:
            # 1. If a data model is deployed through GraphQL, the implements for a child view are sorted
            #   from parent, grandparent, etc.
            # 2. If the grand parent has a direct relation that the parent overwrites to update the source.
            #   The child will get the grandparent's source, not the parent's.
            # We sort the implements in topological order to ensure that the child view get the order grandparent,
            # parent, such that the parent's source is used.
            try:
                dumped["implements"] = [view_id.dump() for view_id in self.topological_sort(resource.implements)]
            except ToolkitCycleError as e:
                warning = MediumSeverityWarning(f"Failed to sort implements for view {resource.as_id()}: {e}")
                warning.print_warning(console=self.console)

        local_properties = local.get("properties", {})
        for prop_id, prop in dumped.get("properties", {}).items():
            if prop_id not in local_properties:
                continue
            local_prop = local_properties[prop_id]
            if all(isinstance(v.get("container"), dict) for v in [prop, local_prop]):
                if prop["container"].get("type") == "container" and "type" not in local_prop["container"]:
                    prop["container"].pop("type", None)
            is_connection_prop = "connectionType" in prop
            is_local_connection_prop = "connectionType" in local_prop
            if (
                is_connection_prop
                and is_local_connection_prop
                and "direction" not in local_prop
                and prop.get("direction") == "outwards"
            ):
                # The API will set the direction to outwards by default, so we remove it from the dump.
                prop.pop("direction", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("implements",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[ViewApply]) -> ViewList:
        try:
            return self.client.data_modeling.views.apply(items)
        except CogniteAPIError as e1:
            if self._is_auto_retryable(e1):
                # Fallback to creating one by one if the error is auto-retryable.
                return self._fallback_create_one_by_one(items, e1)
            elif self._is_false_not_exists(e1, {item.as_id() for item in items}):
                return self._try_to_recover_coupled(items, e1)
            raise

    @staticmethod
    def _is_false_not_exists(e: CogniteAPIError, request_views: set[ViewId]) -> bool:
        if "not exist" not in e.message and 400 <= e.code < 500:
            return False
        results = re.findall(r"'([a-zA-Z0-9_-]+):([a-zA-Z0-9_]+)/([.a-zA-Z0-9_-]+)'", e.message)
        if not results:
            # No view references in the message
            return False
        error_message_views = {ViewId(*result) for result in results}
        return error_message_views.issubset(request_views)

    def _try_to_recover_coupled(self, items: Sequence[ViewApply], original_error: CogniteAPIError) -> ViewList:
        """The /models/views endpoint can give faulty 400 about missing views that are part of the request.

        This method tries to recover from such errors by identifying the strongly connected components in the graph
        defined by the implements and through properties of the views. We then create the components in topological
        order.

        Args:
            items: The items that failed to create.
            original_error: The original error that was raised. If the problem is not recoverable, this error is raised.

        Returns:
            The views that were created.

        """
        views_by_id = {self.get_id(item): item for item in items}
        parents_by_child: dict[ViewId, set[ViewId]] = {
            view_id: {parent for parent in view.implements or [] if parent in views_by_id}
            for view_id, view in views_by_id.items()
        }
        # Check for cycles in the implements graph
        try:
            TopologicalSorter(parents_by_child).static_order()
        except CycleError as e:
            raise ToolkitCycleError(f"Failed to deploy views. This likely due to a cycle in implements. {e.args[1]}")

        dependencies_by_id: dict[ViewId, set[ViewId]] = defaultdict(set)
        for view_id, view in views_by_id.items():
            dependencies_by_id[view_id].update([parent for parent in view.implements or [] if parent in views_by_id])
            for properties in (view.properties or {}).values():
                if isinstance(properties, ReverseDirectRelationApply):
                    if isinstance(properties.through.source, ViewId) and properties.through.source in views_by_id:
                        dependencies_by_id[view_id].add(properties.through.source)

        LowSeverityWarning(
            f"Failed to create {len(items)} views: {escape(original_error.message)}.\nAttempting to recover..."
        ).print_warning(include_timestamp=True, console=self.console)
        created = ViewList([])
        for strongly_connected in tarjan(dependencies_by_id):
            to_create = [views_by_id[view_id] for view_id in strongly_connected]
            try:
                created_set = self.client.data_modeling.views.apply(to_create)
            except CogniteAPIError as error:
                self.client.data_modeling.views.delete(created.as_ids())
                HighSeverityWarning(
                    f"Recovering attempt failed. Could not create views {self.get_ids(to_create)}: "
                    f"{escape(error.message)}.\n Raising original error."
                ).print_warning(console=self.console)
                raise original_error
            created.extend(created_set)
        message = f"Recovery attempt succeeded. Created {len(created)} views."
        if self.console:
            self.console.print(message)
        else:
            print(message)
        return created

    @staticmethod
    def _is_auto_retryable(e: CogniteAPIError) -> bool:
        return isinstance(e.extra, dict) and "isAutoRetryable" in e.extra and e.extra["isAutoRetryable"]

    def _fallback_create_one_by_one(
        self, items: Sequence[ViewApply], e1: CogniteAPIError, warn: bool = True
    ) -> ViewList:
        if warn:
            MediumSeverityWarning(
                f"Failed to create {len(items)} views error:\n{escape(str(e1))}\n\n----------------------------\nTrying to create one by one..."
            ).print_warning(include_timestamp=True, console=self.console)
        created_list = ViewList([])
        for no, item in enumerate(items):
            try:
                created = self.client.data_modeling.views.apply(item)
            except CogniteAPIError as e2:
                e2.failed.extend(items[no + 1 :])
                e2.successful.extend(created_list)
                raise e2 from e1
            else:
                created_list.append(created)
        return created_list

    def retrieve(self, ids: SequenceNotStr[ViewId]) -> ViewList:
        return self.client.data_modeling.views.retrieve(
            cast(Sequence, ids), include_inherited_properties=False, all_versions=False
        )

    def update(self, items: Sequence[ViewApply]) -> ViewList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[ViewId]) -> int:
        to_delete = list(ids)
        nr_of_deleted = 0
        attempt_count = 5
        for attempt_no in range(attempt_count):
            deleted = self.client.data_modeling.views.delete(to_delete)
            nr_of_deleted += len(deleted)
            existing = self.client.data_modeling.views.retrieve(to_delete).as_ids()
            if not existing:
                return nr_of_deleted
            sleep(2)
            to_delete = existing
        else:
            msg = f"  [bold yellow]WARNING:[/] Could not delete views {to_delete} after {attempt_count} attempts."
            if self.console:
                self.console.print(msg)
            else:
                print(msg)
        return nr_of_deleted

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[View]:
        return iter(self.client.data_modeling.views(space=space))

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # The Filter class in the SDK class View implementation is deviating from the API.
        # So we need to modify the spec to match the API.
        parameter_path = ("filter",)
        length = len(parameter_path)
        for item in spec:
            if len(item.path) >= length + 1 and item.path[:length] == parameter_path[:length]:
                # Add extra ANY_STR layer
                # The spec class is immutable, so we use this trick to modify it.
                is_has_data_filter = item.path[1] in ["containers", "views"]
                if is_has_data_filter:
                    # Special handling of the HasData filter that deviates in SDK implementation from API Spec.
                    object.__setattr__(item, "path", item.path[:length] + (ANY_STR,) + item.path[length + 1 :])
                else:
                    object.__setattr__(item, "path", item.path[:length] + (ANY_STR,) + item.path[length:])

        spec.add(ParameterSpec(("filter", ANY_STR), frozenset({"dict"}), is_required=False, _is_nullable=False))
        # The following types are used by the SDK to load the correct class. They are not part of the init,
        # so we need to add it manually.
        spec.update(
            ParameterSpecSet(
                {
                    ParameterSpec(
                        ("implements", ANY_INT, "type"), frozenset({"str"}), is_required=True, _is_nullable=False
                    ),
                    ParameterSpec(
                        ("properties", ANY_STR, "connectionType"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("properties", ANY_STR, "source", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("properties", ANY_STR, "container", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("properties", ANY_STR, "edgeSource", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        ("properties", ANY_STR, "through", "source", "type"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    ParameterSpec(
                        # In the SDK, this is called "property"
                        ("properties", ANY_STR, "through", "identifier"),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                    # Filters are complex, so we do not attempt to give any more specific spec.
                    ParameterSpec(
                        ("filter", ANYTHING),
                        frozenset({"str"}),
                        is_required=True,
                        _is_nullable=False,
                    ),
                }
            )
        )
        spec.discard(
            ParameterSpec(
                # The API spec calls this "identifier", while the SDK calls it "property".
                ("properties", ANY_STR, "through", "property"),
                frozenset({"str"}),
                is_required=True,
                _is_nullable=False,
            )
        )
        return spec

    @classmethod
    def as_str(cls, id: ViewId) -> str:
        return to_directory_compatible(id.external_id)

    def _lookup_views(self, view_ids: list[ViewId]) -> dict[ViewId, View]:
        """Looks up views by their IDs and caches them."""
        missing_ids = [view_id for view_id in view_ids if view_id not in self._view_by_id]
        if missing_ids:
            retrieved_views = self.client.data_modeling.views.retrieve(missing_ids, all_versions=False)
            for view in retrieved_views:
                self._view_by_id[view.as_id()] = view
        return {view_id: self._view_by_id[view_id] for view_id in view_ids if view_id in self._view_by_id}

    def topological_sort(self, view_ids: list[ViewId]) -> list[ViewId]:
        """Sorts the views in topological order based on their implements and through properties."""
        view_by_ids = self._lookup_views(view_ids)
        parents_by_child: dict[ViewId, set[ViewId]] = {}
        for child, view in view_by_ids.items():
            parents_by_child[child] = set()
            for parent in view.implements or []:
                parents_by_child[child].add(parent)
        try:
            sorted_views = list(TopologicalSorter(parents_by_child).static_order())
        except CycleError as e:
            raise ToolkitCycleError(
                f"Failed to sort views topologically. This likely due to a cycle in implements. {e.args[1]}"
            )

        return sorted_views


@final
class DataModelLoader(ResourceLoader[DataModelId, DataModelApply, DataModel, DataModelApplyList, DataModelList]):
    folder_name = "data_models"
    filename_pattern = r"^.*datamodel$"
    resource_cls = DataModel
    resource_write_cls = DataModelApply
    list_cls = DataModelList
    list_write_cls = DataModelApplyList
    kind = "DataModel"
    dependencies = frozenset({SpaceLoader, ViewLoader})
    _doc_url = "Data-models/operation/createDataModels"

    @property
    def display_name(self) -> str:
        return "data models"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataModelApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]

        scope = (
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All()
        )

        return DataModelsAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: DataModelApply | DataModel | dict) -> DataModelId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return DataModelId(space=item["space"], external_id=item["externalId"], version=str(item["version"]))
        return DataModelId(item.space, item.external_id, str(item.version))

    @classmethod
    def dump_id(cls, id: DataModelId) -> dict[str, Any]:
        return id.dump(include_type=False)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]
        for view in item.get("views", []):
            if in_dict(("space", "externalId"), view):
                yield (
                    ViewLoader,
                    ViewId(view["space"], view["externalId"], str(v) if (v := view.get("version")) else None),
                )

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def dump_resource(self, resource: DataModel, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if "views" not in dumped:
            return dumped
        # Sorting in the same order as the local file.
        view_order_by_id = {ViewId.load(v): no for no, v in enumerate(local.get("views", []))}
        end_of_list = len(view_order_by_id)
        dumped["views"] = sorted(dumped["views"], key=lambda v: view_order_by_id.get(ViewId.load(v), end_of_list))
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("views",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: DataModelApplyList) -> DataModelList:
        return self.client.data_modeling.data_models.apply(items)

    def retrieve(self, ids: SequenceNotStr[DataModelId]) -> DataModelList:
        return self.client.data_modeling.data_models.retrieve(cast(Sequence, ids))

    def update(self, items: DataModelApplyList) -> DataModelList:
        updated = self.create(items)
        # There is a bug in the API not raising an exception if view is removed from a data model.
        # So we check here that the update was fixed.
        updated_by_id = {item.as_id(): item for item in updated}
        for local in items:
            item_id = local.as_id()
            if item_id in updated_by_id:
                views_updated = {v.as_id() if isinstance(v, View) else v for v in updated_by_id[item_id].views or []}
                views_local = set(v.as_id() if isinstance(v, ViewApply) else v for v in local.views or [])
                missing = views_local - views_updated
                extra = views_updated - views_local
                if missing or extra:
                    raise CogniteAPIError(
                        f"The API did not update the data model, {item_id} correctly. You might have "
                        f"to increase the version number of the data model for it to update.\nMissing views in CDF: {missing}\n"
                        f"Extra views in the CDF: {extra}",
                        code=500,
                    )
            else:
                raise CogniteAPIError(
                    f"The data model {item_id} was not updated. Please check the data model manually.",
                    code=500,
                )

        return updated

    def delete(self, ids: SequenceNotStr[DataModelId]) -> int:
        return len(self.client.data_modeling.data_models.delete(cast(Sequence, ids)))

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[DataModel]:
        return iter(self.client.data_modeling.data_models(space=space, include_global=False))

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # ViewIds have the type set in the API Spec, but this is hidden in the SDK classes,
        # so we need to add it manually.
        spec.add(ParameterSpec(("views", ANY_INT, "type"), frozenset({"str"}), is_required=True, _is_nullable=False))
        return spec

    @classmethod
    def as_str(cls, id: DataModelId) -> str:
        return to_directory_compatible(id.external_id)


@final
class NodeLoader(ResourceContainerLoader[NodeId, NodeApply, Node, NodeApplyList, NodeList]):
    item_name = "nodes"
    folder_name = "data_models"
    filename_pattern = r"^.*node$"
    resource_cls = Node
    resource_write_cls = NodeApply
    list_cls = NodeList
    list_write_cls = NodeApplyList
    kind = "Node"
    dependencies = frozenset({SpaceLoader, ViewLoader, ContainerLoader})
    _doc_url = "Instances/operation/applyNodeAndEdges"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None = None,
        view_id: ViewId | None = None,
    ) -> None:
        super().__init__(client, build_dir, console)
        # View ID is used to retrieve nodes with properties.
        self.view_id = view_id

    @property
    def display_name(self) -> str:
        return "nodes"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[NodeApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(
            actions,
            DataModelInstancesAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelInstancesAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: NodeApply | Node | dict) -> NodeId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return NodeId(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: NodeId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]
        for source in item.get("sources", []):
            if (identifier := source.get("source")) and isinstance(identifier, dict):
                if identifier.get("type") == "view" and in_dict(("space", "externalId", "version"), identifier):
                    yield (
                        ViewLoader,
                        ViewId(
                            identifier["space"],
                            identifier["externalId"],
                            str(v) if (v := identifier.get("version")) else None,
                        ),
                    )
                elif identifier.get("type") == "container" and in_dict(("space", "externalId"), identifier):
                    yield ContainerLoader, ContainerId(identifier["space"], identifier["externalId"])

    def dump_resource(self, resource: Node, local: dict[str, Any] | None = None) -> dict[str, Any]:
        # CDF resource does not have properties set, so we need to do a lookup
        local = local or {}
        sources = [ViewId.load(source["source"]) for source in local.get("sources", []) if "source" in source]

        if sources:
            try:
                res = self.client.data_modeling.instances.retrieve(nodes=resource.as_id(), sources=sources)
            except CogniteAPIError:
                # View does not exist
                dumped = resource.as_write().dump()
            else:
                dumped = res.nodes[0].as_write().dump() if len(res.nodes) > 0 else resource.as_write().dump()
        else:
            dumped = resource.as_write().dump()

        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)

        if "instanceType" in dumped and "instanceType" not in local:
            # Toolkit uses file suffix to determine instanceType, so we need to remove it from the CDF resource
            # to match the local resource.
            dumped.pop("instanceType")

        return dumped

    def create(self, items: NodeApplyList) -> NodeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            # Note replace should never be relevant as Toolkit always checks whether the node exists before applying.
            nodes=items,
            auto_create_direct_relations=True,
            replace=True,
        )
        return result.nodes

    def retrieve(self, ids: SequenceNotStr[NodeId]) -> NodeList:
        return self.client.data_modeling.instances.retrieve(nodes=cast(Sequence, ids), sources=self.view_id).nodes

    def update(self, items: NodeApplyList) -> NodeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            nodes=items, auto_create_direct_relations=True, replace=False
        )
        return result.nodes

    def delete(self, ids: SequenceNotStr[NodeId]) -> int:
        try:
            deleted = self.client.data_modeling.instances.delete(nodes=cast(Sequence, ids))
        except CogniteAPIError as e:
            if "not exist" in e.message and "space" in e.message.lower():
                return 0
            raise e
        return len(deleted.nodes)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Node]:
        return iter(
            iterate_instances(self.client, space=space, instance_type="node", source=self.view_id, console=self.console)
        )

    def count(self, ids: SequenceNotStr[NodeId]) -> int:
        return len(ids)

    def drop_data(self, ids: SequenceNotStr[NodeId]) -> int:
        # Nodes will be deleted in .delete call.
        return 0

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        node_spec = super().get_write_cls_parameter_spec()
        # This is a deviation between the SDK and the API
        node_spec.add(ParameterSpec(("instanceType",), frozenset({"str"}), is_required=False, _is_nullable=False))
        node_spec.add(
            ParameterSpec(
                ("sources", ANY_INT, "source", "type"),
                frozenset({"str"}),
                is_required=True,
                _is_nullable=False,
            )
        )
        return ParameterSpecSet(node_spec, spec_name=cls.__name__)

    @classmethod
    def as_str(cls, id: NodeId) -> str:
        return to_directory_compatible(f"{id.space}_{id.external_id}")


class GraphQLLoader(
    ResourceContainerLoader[
        DataModelId, GraphQLDataModelWrite, GraphQLDataModel, GraphQLDataModelWriteList, GraphQLDataModelList
    ]
):
    folder_name = "data_models"
    filename_pattern = r"^.*GraphQLSchema"
    resource_cls = GraphQLDataModel
    resource_write_cls = GraphQLDataModelWrite
    list_cls = GraphQLDataModelList
    list_write_cls = GraphQLDataModelWriteList
    kind = "GraphQLSchema"
    dependencies = frozenset({SpaceLoader, ContainerLoader})
    item_name = "views"
    _doc_url = "Data-models/operation/createDataModels"
    _hash_name = "CDFToolkitHash:"

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None) -> None:
        super().__init__(client, build_dir, console)
        self._graphql_filepath_cache: dict[DataModelId, Path] = {}
        self._datamodels_by_view_id: dict[ViewId, set[DataModelId]] = defaultdict(set)
        self._dependencies_by_datamodel_id: dict[DataModelId, set[ViewId | DataModelId]] = {}

    @property
    def display_name(self) -> str:
        return "graph QL schemas"

    @classmethod
    def get_id(cls, item: GraphQLDataModelWrite | GraphQLDataModel | dict) -> DataModelId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId", "version"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return DataModelId(space=item["space"], external_id=item["externalId"], version=str(item["version"]))
        return DataModelId(item.space, item.external_id, str(item.version))

    @classmethod
    def dump_id(cls, id: DataModelId) -> dict[str, Any]:
        return id.dump(include_type=False)

    @classmethod
    def get_required_capability(
        cls, items: Sequence[GraphQLDataModelWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = [DataModelsAcl.Action.Read] if read_only else [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write]
        return DataModelsAcl(
            actions,
            DataModelsAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelsAcl.Scope.All(),
        )

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        raw_yaml = load_yaml_inject_variables(
            self.safe_read(filepath),
            environment_variables or {},
            original_filepath=filepath,
        )
        raw_list = raw_yaml if isinstance(raw_yaml, list) else [raw_yaml]

        for item in raw_list:
            model_id = self.get_id(item)
            # Find the GraphQL files adjacent to the DML files
            graphql_file = filepath.with_suffix(".graphql")
            if not graphql_file.is_file():
                raise ToolkitFileNotFoundError(
                    f"Failed to find GraphQL file. Expected {graphql_file.name} adjacent to {filepath.as_posix()}"
                )

            self._graphql_filepath_cache[model_id] = graphql_file
            graphql_content = safe_read(graphql_file, encoding=BUILD_FOLDER_ENCODING)

            parser = GraphQLParser(graphql_content, model_id)
            try:
                for view in parser.get_views():
                    self._datamodels_by_view_id[view].add(model_id)
                self._dependencies_by_datamodel_id[model_id] = parser.get_dependencies()
            except Exception as e:
                # We catch a broad exception here to give a more user-friendly error message.
                raise GraphQLParseError(f"Failed to parse GraphQL file {graphql_file.as_posix()}: {e}") from e

            # Add hash to description
            description = item.get("description", "")
            hash_ = calculate_hash(graphql_content)[:8]
            suffix = f"{self._hash_name}{hash_}"
            if len(description) + len(suffix) > 1024:
                LowSeverityWarning(f"Description is above limit for {model_id}. Truncating...").print_warning()
                description = description[: 1024 - len(suffix) + 1 - 3] + "..."
            description += f" {suffix}"
            item["description"] = description
            item["graphqlFile"] = hash_
        return raw_list

    def dump_resource(self, resource: GraphQLDataModel, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        for key in ["dml", "preserveDml"]:
            # Local values that are not returned from the API
            if key in local:
                dumped[key] = local[key]

        description = resource.description or ""
        if match := re.match(rf"(.|\n)*( {self._hash_name}([a-f0-9]{{8}}))$", description):
            dumped["graphqlFile"] = match.group(3)
        return dumped

    def create(self, items: GraphQLDataModelWriteList) -> list[DMLApplyResult]:
        creation_order = self._topological_sort(items)

        created_list: list[DMLApplyResult] = []
        for item in creation_order:
            item_id = item.as_id()
            graphql_file_content = self._get_graphql_content(item_id)
            if "--verbose" in sys.argv:
                print(f"Deploying GraphQL schema {item_id}")

            created = self.client.dml.apply_dml(
                item.as_id(),
                dml=graphql_file_content,
                name=item.name,
                description=item.description,
                previous_version=item.previous_version,
                preserve_dml=item.preserve_dml,
            )
            created_list.append(created)
        return created_list

    def _get_graphql_content(self, data_model_id: DataModelId) -> str:
        filepath = self._graphql_filepath_cache.get(data_model_id)
        if filepath is None:
            raise ToolkitFileNotFoundError(f"Could not find the GraphQL file for {data_model_id}")
        return safe_read(filepath)

    def retrieve(self, ids: SequenceNotStr[DataModelId]) -> GraphQLDataModelList:
        result = self.client.data_modeling.data_models.retrieve(list(ids), inline_views=False)
        return GraphQLDataModelList([GraphQLDataModel._load(d.dump()) for d in result])

    def update(self, items: GraphQLDataModelWriteList) -> list[DMLApplyResult]:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[DataModelId]) -> int:
        retrieved = self.retrieve(ids)
        views = {view for dml in retrieved for view in dml.views or []}
        deleted = len(self.client.data_modeling.data_models.delete(list(ids)))
        deleted += len(self.client.data_modeling.views.delete(list(views)))
        return deleted

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[GraphQLDataModel]:
        return iter(GraphQLDataModel._load(d.dump()) for d in self.client.data_modeling.data_models)

    def count(self, ids: SequenceNotStr[DataModelId]) -> int:
        retrieved = self.retrieve(ids)
        return sum(len(d.views or []) for d in retrieved)

    def drop_data(self, ids: SequenceNotStr[DataModelId]) -> int:
        return self.delete(ids)

    def _topological_sort(self, items: GraphQLDataModelWriteList) -> list[GraphQLDataModelWrite]:
        to_sort = {item.as_id(): item for item in items}
        dependencies: dict[DataModelId, set[DataModelId]] = {}
        for item in items:
            item_id = item.as_id()
            dependencies[item_id] = set()
            for dependency in self._dependencies_by_datamodel_id.get(item_id, []):
                if isinstance(dependency, DataModelId) and dependency in to_sort:
                    dependencies[item_id].add(dependency)
                elif isinstance(dependency, ViewId):
                    for model_id in self._datamodels_by_view_id.get(dependency, set()):
                        if model_id in to_sort:
                            dependencies[item_id].add(model_id)
        try:
            return [to_sort[item_id] for item_id in TopologicalSorter(dependencies).static_order()]
        except CycleError as e:
            raise ToolkitCycleError(
                f"Cannot create GraphQL schemas. Cycle detected between models {e.args} using the @import directive.",
                *e.args[1:],
            )


@final
class EdgeLoader(ResourceContainerLoader[EdgeId, EdgeApply, Edge, EdgeApplyList, EdgeList]):
    item_name = "edges"
    folder_name = "data_models"
    filename_pattern = r"^.*edge"
    resource_cls = Edge
    resource_write_cls = EdgeApply
    list_cls = EdgeList
    list_write_cls = EdgeApplyList
    kind = "Edge"
    dependencies = frozenset({SpaceLoader, ViewLoader, ContainerLoader, NodeLoader})
    _doc_url = "Instances/operation/applyNodeAndEdges"

    @property
    def display_name(self) -> str:
        return "edges"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[EdgeApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(
            actions,
            DataModelInstancesAcl.Scope.SpaceID(list({item.space for item in items}))
            if items is not None
            else DataModelInstancesAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: EdgeApply | Edge | dict) -> EdgeId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"space", "externalId"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return EdgeId(space=item["space"], external_id=item["externalId"])
        return item.as_id()

    @classmethod
    def dump_id(cls, id: EdgeId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "space" in item:
            yield SpaceLoader, item["space"]
        for source in item.get("sources", []):
            if (identifier := source.get("source")) and isinstance(identifier, dict):
                if identifier.get("type") == "view" and in_dict(("space", "externalId", "version"), identifier):
                    yield (
                        ViewLoader,
                        ViewId(
                            identifier["space"],
                            identifier["externalId"],
                            str(v) if (v := identifier.get("version")) else None,
                        ),
                    )
                elif identifier.get("type") == "container" and in_dict(("space", "externalId"), identifier):
                    yield ContainerLoader, ContainerId(identifier["space"], identifier["externalId"])

        for key in ["startNode", "endNode", "type"]:
            if node_ref := item.get(key):
                if isinstance(node_ref, dict) and in_dict(("space", "externalId"), node_ref):
                    yield NodeLoader, NodeId(node_ref["space"], node_ref["externalId"])

    def dump_resource(self, resource: Edge, local: dict[str, Any] | None = None) -> dict[str, Any]:
        # CDF resource does not have properties set, so we need to do a lookup
        local = local or {}
        sources = [ViewId.load(source["source"]) for source in local.get("sources", []) if "source" in source]
        try:
            cdf_resource_with_properties = self.client.data_modeling.instances.retrieve(
                edges=resource.as_id(), sources=sources
            ).edges[0]
        except (CogniteAPIError, IndexError):
            # View or Edge does not exist
            dumped = resource.as_write().dump()
        else:
            dumped = cdf_resource_with_properties.as_write().dump()

        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        if dumped.get("instanceType") == "edge" and "instanceType" not in local:
            # Toolkit uses file suffix to determine instanceType, so we need to remove it from the CDF resource
            # to match the local resource.
            dumped.pop("instanceType", None)

        return dumped

    def create(self, items: EdgeApplyList) -> EdgeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            edges=items, auto_create_direct_relations=True, replace=False
        )
        return result.edges

    def retrieve(self, ids: SequenceNotStr[EdgeId]) -> EdgeList:
        return self.client.data_modeling.instances.retrieve(nodes=cast(Sequence, ids)).edges

    def update(self, items: EdgeApplyList) -> EdgeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            edges=items, auto_create_direct_relations=False, replace=True
        )
        return result.edges

    def delete(self, ids: SequenceNotStr[EdgeId]) -> int:
        try:
            deleted = self.client.data_modeling.instances.delete(edges=cast(Sequence, ids))
        except CogniteAPIError as e:
            if "not exist" in e.message and "space" in e.message.lower():
                return 0
            raise e
        return len(deleted.edges)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Edge]:
        return iter(iterate_instances(self.client, space=space, instance_type="edge"))

    def count(self, ids: SequenceNotStr[EdgeId]) -> int:
        return len(ids)

    def drop_data(self, ids: SequenceNotStr[EdgeId]) -> int:
        # Edges will be deleted in .delete call.
        return 0

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        node_spec = super().get_write_cls_parameter_spec()
        # This is a deviation between the SDK and the API
        node_spec.add(ParameterSpec(("instanceType",), frozenset({"str"}), is_required=False, _is_nullable=False))
        node_spec.add(
            ParameterSpec(
                ("sources", ANY_INT, "source", "type"),
                frozenset({"str"}),
                is_required=True,
                _is_nullable=False,
            )
        )
        return ParameterSpecSet(node_spec, spec_name=cls.__name__)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("sources",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: dm_identifier(x["source"]))
        return super().diff_list(local, cdf, json_path)



from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Any, Optional

import pytz  # type: ignore
from cognite.client import CogniteClient
from common.dataclass.common import Vec3f

logger = logging.getLogger(__name__)


def create_apm_observation_from_reading(
    client: CogniteClient,
    file_external_id: str,
    timeseries_external_id: str,
    value: float,
    timestamp: int,
    pose: Vec3f,
    message: str,
    apm_checklistitem_external_id: Optional[str] = None,
):
    """Create an APM observation from a gauge reading.

    NOTE: this is a preliminary implementation, where we're not yet using the
    gql pygen SDK made by Cognite (which will make this a lot cleaner)

    """
    apm_config = _get_apm_config(client=client, config_external_id="inrobot")
    if apm_config is None:
        logger.error("Could not find APM config")
        return

    space_id = str(apm_config.get("appDataSpaceId"))
    appdata_space_version = str(apm_config.get("appDataSpaceVersion"))

    views = _get_views(
        client=client,
        space=space_id,
        datamodel_id=space_id,
        datamodel_version=appdata_space_version,
        view_ids=["APM_Measurement", "Vec3f", "APM_Observation"],
    )

    def _get_view_version(views: list[dict[str, Any]], view_id: str) -> Optional[str]:
        view = next((item for item in views if item["externalId"] == view_id), None)
        if view is None:
            raise ValueError(f"Could not find view {view_id}")
        return view.get("version")

    # create the APM measurement
    if isinstance(timestamp, int):
        timestamp_s = round(timestamp / 1000)
        timestamp_dt = datetime.fromtimestamp(float(timestamp_s), pytz.UTC)
    elif isinstance(timestamp, datetime):
        timestamp_dt = timestamp
    else:
        raise ValueError(f"Timestamp is of type {type(timestamp)}, needs to be either `int` or `datetime`.")

    apm_measurement_external_id = f"measurement_{timeseries_external_id}_{timestamp}"
    apm_measurement = {
        "timeSeries": timeseries_external_id,
        "measuredAt": timestamp_dt.isoformat(),
        "numericValue": value,
    }

    vec3f_external_id = f"position_{timeseries_external_id}_{timestamp}"
    vec3f = {"x": pose.x, "y": pose.y, "z": pose.z}

    apm_observation_external_id = f"observation_{timeseries_external_id}_{timestamp}"
    apm_observation = {
        "fileIds": [file_external_id],
        "position": {"externalId": vec3f_external_id, "space": space_id},
        "description": "Gauge reading",
        "note": message,
        "createdById": "CDF Gauge Reader Function",
        "updatedById": "CDF Gauge Reader Function",
    }

    # edge apm_observation -> apm_measurement
    edge_apm_observation_apm_measurement = {
        "instanceType": "edge",
        "space": space_id,
        "externalId": f"{apm_observation_external_id}.measurements_{apm_measurement_external_id}",
        "startNode": {"externalId": apm_observation_external_id, "space": space_id},
        "endNode": {"externalId": apm_measurement_external_id, "space": space_id},
        "type": {"externalId": "APM_Observation.measurements", "space": space_id},
    }

    # assemble all instances (nodes and edges)
    instances = [
        {
            "instanceType": "node",
            "space": space_id,
            "externalId": apm_measurement_external_id,
            "sources": [
                {
                    "source": {
                        "type": "view",
                        "space": space_id,
                        "externalId": "APM_Measurement",
                        "version": _get_view_version(views, "APM_Measurement"),
                    },
                    "properties": apm_measurement,
                }
            ],
        },
        {
            "instanceType": "node",
            "space": space_id,
            "externalId": vec3f_external_id,
            "sources": [
                {
                    "source": {
                        "type": "view",
                        "space": space_id,
                        "externalId": "Vec3f",
                        "version": _get_view_version(views, "Vec3f"),
                    },
                    "properties": vec3f,
                }
            ],
        },
        {
            "instanceType": "node",
            "space": space_id,
            "externalId": apm_observation_external_id,
            "sources": [
                {
                    "source": {
                        "type": "view",
                        "space": space_id,
                        "externalId": "APM_Observation",
                        "version": _get_view_version(views, "APM_Observation"),
                    },
                    "properties": apm_observation,
                }
            ],
        },
        edge_apm_observation_apm_measurement,
    ]

    if apm_checklistitem_external_id is not None:
        edge_apm_checklist_apm_observation = {
            "instanceType": "edge",
            "space": space_id,
            "externalId": f"{apm_checklistitem_external_id}.observations_{apm_observation_external_id}",
            "startNode": {"externalId": apm_checklistitem_external_id, "space": space_id},
            "endNode": {"externalId": apm_observation_external_id, "space": space_id},
            "type": {"externalId": "APM_ChecklistItem.observations", "space": space_id},
        }
        instances.append(edge_apm_checklist_apm_observation)

    logger.debug(f"Creating instances request: {json.dumps(instances, indent=2)}")

    response = client.post(
        f"/api/v1/projects/{client.config.project}/models/instances", json={"items": instances}
    ).json()

    logger.debug(f"Creating instances response: {json.dumps(response, indent=2)}")


def _get_views(
    client: CogniteClient, space: str, datamodel_id: str, datamodel_version: Optional[str] = None, view_ids=list[str]
) -> list[dict[str, Any]]:
    response = client.post(
        f"/api/v1/projects/{client.config.project}/models/datamodels/byids",
        json={
            "items": [{"space": space, "externalId": datamodel_id, "version": datamodel_version}],
        },
    ).json()

    # get latest datamodel, if no version is specified
    datamodel = max(response["items"], key=lambda item: item["createdTime"], default=None)
    if datamodel is None:
        raise ValueError(f"No datamodels with id {datamodel_id} found with version {datamodel_version}.")

    views = datamodel.get("views")
    if views is None or views == []:
        raise ValueError(f"Datamodel {datamodel_id} has no views.")

    available_views = {view.get("externalId"): view.get("version") for view in views}

    items = [
        {"space": space, "externalId": requested_view_id, "version": available_views.get(requested_view_id)}
        for requested_view_id in view_ids
        if available_views.get(requested_view_id) is not None
    ]

    if len(items) != len(view_ids):
        missing_views = set(view_ids) - set([item["externalId"] for item in items])
        raise ValueError(f"Views {', '.join(missing_views)} not found in datamodel {datamodel_id}.")

    views = client.post(
        f"/api/v1/projects/{client.config.project}/models/views/byids",
        json={
            "items": items,
        },
    ).json()

    return views["items"]


def _get_apm_config(client: CogniteClient, config_external_id: str) -> Optional[dict[str, Any]]:
    space = "APM_Config"
    datamodel_id = "APM_Config"
    view_id = "APM_Config"
    instance_external_ids = [config_external_id, "default-config"]

    views = _get_views(client=client, space=space, datamodel_id=datamodel_id, view_ids=[view_id])

    if len(views) != 1:
        raise ValueError(f"Expected to find exactly one view for {view_id}, found {len(views)}.")

    view = views[0]

    response = client.post(
        f"/api/v1/projects/{client.config.project}/models/instances/byids",
        json={
            "items": [
                {"instanceType": "node", "externalId": external_id, "space": space}
                for external_id in instance_external_ids
            ],
            "sources": [
                {
                    "source": {
                        "type": "view",
                        "space": space,
                        "externalId": view.get("externalId"),
                        "version": view.get("version"),
                    }
                }
            ],
        },
    ).json()

    if response.get("items") == []:
        logger.error("Could not find APM config, not upserting any APM_Observations now.")
        return None

    # Try to find the first item with "externalId" == config_external_id
    apm_config = next((item for item in response["items"] if item["externalId"] == config_external_id), None)

    # If no such item is found, try to find the first item with "externalId" == "default"
    if apm_config is None:
        apm_config = next((item for item in response["items"] if item["externalId"] == "default-config"), None)
        if apm_config is None:
            raise ValueError(f"Could not find APM config with externalId {config_external_id} or `default-config`.")

    apm_config = apm_config.get("properties").get(space).get(f"{view.get('externalId')}/{view.get('version')}")

    return apm_config


def _get_position_from_metadata_to_vec3f(metadata: dict[str, Any]) -> Vec3f:
    """Get position from metadata."""
    x = metadata.get("waypoint_tform_body_x")
    y = metadata.get("waypoint_tform_body_x")
    z = metadata.get("waypoint_tform_body_x")

    if not x or not y or not z:
        raise ValueError(
            f"Missing metadata field. Required metadata fields \
            are waypoint_tform_body_x, waypoint_tform_body_x, waypoint_tform_body_x. File metadata keys: {metadata.keys()}"
        )

    return Vec3f(x=float(x), y=float(y), z=float(z))



from concurrent.futures import ThreadPoolExecutor

from cognite.client.utils._concurrency import ConcurrencySettings, TaskExecutor

from cognite_toolkit._cdf_tk.client._constants import DATA_MODELING_MAX_WRITE_WORKERS

_DATA_MODELING_WRITE_THREAD_POOL_EXECUTOR_SINGLETON: ThreadPoolExecutor


class ToolkitConcurrencySettings(ConcurrencySettings):
    """
    Toolkit-specific concurrency settings for CDF Toolkit operations.
    """

    @classmethod
    def get_data_modeling_write_executor(cls) -> TaskExecutor:
        """
        The data modeling write backend has different concurrency limits compared with the rest of CDF.
        Thus, we use a dedicated executor for these endpoints to match the backend requirements.

        Returns:
            TaskExecutor: The data modeling write executor.
        """
        if cls.uses_mainthread():
            return cls.get_mainthread_executor()

        global _DATA_MODELING_WRITE_THREAD_POOL_EXECUTOR_SINGLETON
        try:
            executor = _DATA_MODELING_WRITE_THREAD_POOL_EXECUTOR_SINGLETON
        except NameError:
            # TPE has not been initialized
            executor = _DATA_MODELING_WRITE_THREAD_POOL_EXECUTOR_SINGLETON = ThreadPoolExecutor(
                DATA_MODELING_MAX_WRITE_WORKERS
            )
        return executor



from __future__ import annotations

from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from typing import Any

from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk._parameters.constants import ANY_INT, ANYTHING
from cognite_toolkit._cdf_tk._parameters.data_classes import ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client.data_classes.agents import Agent, AgentList, AgentWrite, AgentWriteList
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader


class AgentLoader(ResourceLoader[str, AgentWrite, Agent, AgentWriteList, AgentList]):
    folder_name = "agents"
    filename_pattern = r".*\.Agent$"  # Matches all yaml files whose stem ends with '.Agent'.
    resource_cls = Agent
    resource_write_cls = AgentWrite
    list_cls = AgentList
    list_write_cls = AgentWriteList
    kind = "Agent"
    _doc_base_url = ""
    _doc_url = "https://pr-2829.specs.preview.cogniteapp.com/20230101-alpha.json.html"

    @classmethod
    def get_id(cls, item: AgentWrite | Agent | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[AgentWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        return []

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Tool configuration is a dict where the accepted keys depend on the tool type.
        spec.add(
            ParameterSpec(
                ("tools", ANY_INT, "configuration", ANYTHING),
                frozenset({"dict"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        return spec

    def create(self, items: AgentWriteList) -> AgentList:
        return self.client.agents.apply(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> AgentList:
        return self.client.agents.retrieve(ids)

    def update(self, items: AgentWriteList) -> AgentList:
        return self.client.agents.apply(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.agents.delete(ids)
        except CogniteAPIError:
            deleted = 0
            for id in ids:
                try:
                    self.client.agents.delete(id)
                    deleted += 1
                except CogniteAPIError:
                    # accepted because the resource may not exist
                    pass
            return deleted
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Agent]:
        return iter([])



from pydantic import Field, field_validator

from cognite_toolkit._cdf_tk.constants import FORBIDDEN_SPACES, SPACE_FORMAT_PATTERN
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import ToolkitResource


class SpaceYAML(ToolkitResource):
    space: str = Field(
        description="The Space identifier (id).",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    name: str | None = Field(None, description="Name of the space.", max_length=1024)
    description: str | None = Field(None, description="The description of the space.", max_length=255)

    @field_validator("space")
    @classmethod
    def check_forbidden_space_value(cls, val: str) -> str:
        """Check the space name not present in forbidden set"""
        if val in FORBIDDEN_SPACES:
            raise ValueError(f"'{val}' is a reserved space. Reserved Spaces: {humanize_collection(FORBIDDEN_SPACES)}")
        return val



from __future__ import annotations

from typing import Literal, cast

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import CredentialProvider

from .api.agents.agents import AgentsAPI
from .api.dml import DMLAPI
from .api.extended_data_modeling import ExtendedDataModelingAPI
from .api.extended_raw import ExtendedRawAPI
from .api.extended_timeseries import ExtendedTimeSeriesAPI
from .api.location_filters import LocationFiltersAPI
from .api.lookup import LookUpGroup
from .api.robotics import RoboticsAPI
from .api.verify import VerifyAPI


class ToolkitClientConfig(ClientConfig):
    def __init__(
        self,
        client_name: str,
        project: str,
        credentials: CredentialProvider,
        api_subversion: str | None = None,
        base_url: str | None = None,
        max_workers: int | None = None,
        is_strict_validation: bool = True,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
        file_transfer_timeout: int | None = None,
        debug: bool = False,
    ) -> None:
        super().__init__(
            client_name=client_name,
            project=project,
            credentials=credentials,
            api_subversion=api_subversion,
            base_url=base_url,
            max_workers=max_workers,
            headers=headers,
            timeout=timeout,
            file_transfer_timeout=file_transfer_timeout,
            debug=debug,
        )
        self.is_strict_validation = is_strict_validation

    @property
    def cloud_provider(self) -> Literal["azure", "aws", "gcp", "unknown"]:
        cdf_cluster = self.cdf_cluster
        if cdf_cluster is None:
            return "unknown"
        elif cdf_cluster.startswith("az-") or cdf_cluster in {"azure-dev", "bluefield", "westeurope-1"}:
            return "azure"
        elif cdf_cluster.startswith("aws-") or cdf_cluster in {"orangefield"}:
            return "aws"
        elif cdf_cluster.startswith("gc-") or cdf_cluster in {
            "greenfield",
            "asia-northeast1-1",
            "cognitedata-development",
            "cognitedata-production",
        }:
            return "gcp"
        else:
            return "unknown"

    @property
    def is_private_link(self) -> bool:
        if "cognitedata.com" not in self.base_url:
            return False
        subdomain = self.base_url.split("cognitedata.com", maxsplit=1)[0]
        return "plink" in subdomain


class ToolkitClient(CogniteClient):
    def __init__(self, config: ToolkitClientConfig | None = None, enable_set_pending_ids: bool = False) -> None:
        super().__init__(config=config)
        self.location_filters = LocationFiltersAPI(self._config, self._API_VERSION, self)
        self.robotics = RoboticsAPI(self._config, self._API_VERSION, self)
        self.dml = DMLAPI(self._config, self._API_VERSION, self)
        self.verify = VerifyAPI(self._config, self._API_VERSION, self)
        self.lookup = LookUpGroup(self._config, self._API_VERSION, self)
        self.agents = AgentsAPI(self._config, self._API_VERSION, self)
        self.data_modeling: ExtendedDataModelingAPI = ExtendedDataModelingAPI(self._config, self._API_VERSION, self)
        if enable_set_pending_ids:
            self.time_series: ExtendedTimeSeriesAPI = ExtendedTimeSeriesAPI(self._config, self._API_VERSION, self)
        self.raw: ExtendedRawAPI = ExtendedRawAPI(self._config, self._API_VERSION, self)

    @property
    def config(self) -> ToolkitClientConfig:
        """Returns a config object containing the configuration for the current client.

        Returns:
            ToolkitClientConfig: The configuration object.
        """
        return cast(ToolkitClientConfig, self._config)



from typing import TypeVar

T = TypeVar("T")


def tarjan(dependencies_by_id: dict[T, set[T]]) -> list[set[T]]:
    """Returns the strongly connected components of the dependency graph
     in topological order.

    Args:
        dependencies_by_id: A dictionary where the keys are ids and the values are sets of ids that the key depends on.

    Returns:
        A list of sets of ids that are strongly connected components in the dependency graph.
    """

    stack = []
    stack_set = set()
    index: dict[T, int] = {}
    lowlink = {}
    result = []

    def visit(v: T) -> None:
        index[v] = len(index)
        lowlink[v] = index[v]
        stack.append(v)
        stack_set.add(v)
        for w in dependencies_by_id.get(v, []):
            if w not in index:
                visit(w)
                lowlink[v] = min(lowlink[w], lowlink[v])
            elif w in stack_set:
                lowlink[v] = min(lowlink[v], index[w])
        if lowlink[v] == index[v]:
            scc = set()
            dependency: T | None = None
            while v != dependency:
                dependency = stack.pop()
                scc.add(dependency)
                stack_set.remove(dependency)
            result.append(scc)

    for view_id in dependencies_by_id.keys():
        if view_id not in index:
            visit(view_id)
    return result



"""Gauge reader handler."""

from __future__ import annotations

import logging
import sys
from typing import Optional

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadataUpdate, LabelFilter, TimeSeriesUpdate
from cognite.extractorutils.uploader import TimeSeriesUploadQueue
from common.apm_helpers import _get_position_from_metadata_to_vec3f, create_apm_observation_from_reading
from common.cdf_helpers import call_vision_api_endpoint, create_annotations, create_missing_labels

MAX_GET_JOB_ATTEMPTS = 20

METADATA = {
    "dial": {"required": [], "all": ["min_level", "max_level", "dead_angle"]},
    "digital": {
        "required": ["comma_pos"],
        "all": ["comma_pos", "min_num_digits", "max_num_digits"],
    },
    "level": {"required": [], "all": ["min_level", "max_level"]},
    "valve": {"required": [], "all": []},
}
SUPPORTED_GAUGES = ["dial", "digital", "level", "valve"]
GAUGE_FEATURE_TYPES = {
    "dial": "DialGaugeDetection",
    "digital": "DigitalGaugeDetection",
    "level": "LevelGaugeDetection",
    "valve": "ValveDetection",
}

GAUGE_PARAMETERS_TYPES = {
    "dial": "dialGaugeDetectionParameters",
    "digital": "digitalGaugeDetectionParameters",
    "level": "levelGaugeDetectionParameters",
    "valve": "valveDetectionParameters",
}

logger = logging.getLogger(__name__)

BOUNDING_BOX_LABELS = {"dial": "dial-gauge", "digital": "digital-gauge", "level": "level-gauge", "valve": "valve"}
KEYPOINT_LABELS = {
    "dial": "dial-gauge-keypoints",
    "digital": "digital-gauge-keypoints",
    "level": "level-gauge-keypoints",
    "valve": "valve-keypoints",
}


def update_ts_metadata(client, file_metadata, metadata_keys):
    """Update a timeseries with gauge reading metadata."""
    metadata = {key: file_metadata.get(key) for key in metadata_keys}
    logger.info(f"Timeseries {file_metadata['ts_external_id']} updated with metadata {metadata}")
    client.time_series.update(TimeSeriesUpdate(external_id=file_metadata["ts_external_id"]).metadata.add(metadata))


def gauge_reading_attributes_from_response(res: dict, gauge_type: str) -> Optional[dict]:
    """Return the gauge reading annotations from the API response."""
    if not res:
        logger.error("API call failed. Did not get a response.")
        return None
    if not res["items"]:
        logger.error(f"No items in result: {res}")
        return None

    if gauge_type == "valve":
        valve_readings = [annotation for annotation in res["items"][0]["predictions"]["valvePredictions"]]
        if not valve_readings:
            logger.error(f"No annotation of type {gauge_type} in result.")
            return None
        for reading in valve_readings:
            if reading.get("keypointCollection", None):
                data = reading["keypointCollection"]
    else:
        gauge_readings = [annotation for annotation in res["items"][0]["predictions"][gauge_type + "GaugePredictions"]]
        gauge_attributes = []

        for gauge_reading in gauge_readings:
            if gauge_type == "digital":
                if gauge_reading.get("attributes", None):
                    gauge_attributes.append(gauge_reading)
            else:
                if gauge_reading.get("keypointCollection", None):
                    gauge_attributes.append(gauge_reading["keypointCollection"])

        if not gauge_attributes:
            logger.error(f"No annotation of type {gauge_type} in result.")
            return None

        if "attributes" not in gauge_attributes[0]:
            logger.error("No attributes in result.")
            return None
        data = gauge_attributes[0]["attributes"]

    return data


def handle_failed_upload(client: CogniteClient, id: int, error_message: str, data: dict, metadata: dict | None = None):
    """Log error message and update a file that has failed."""
    logger.error(error_message)

    if data["gauge_type"] == "valve":
        client.files.update(
            FileMetadataUpdate(id=id)
            .labels.remove(data["input_label"])
            .labels.add([data["output_label"], data["failed_label"]])
            .metadata.add({"error_message": error_message})
        )
    else:
        client.files.update(
            FileMetadataUpdate(id=id)
            .labels.remove(data["input_label"])
            .labels.add([data["output_label"], data["failed_label"]])
            .metadata.add(
                {"error_message": error_message, **metadata}
                if metadata is not None
                else {"error_message": error_message}
            )
        )


def to_input_metadata(keys: list[str], metadata: dict):
    return {to_camel_case(key): metadata.get(key) for key in keys}


def get_timestamp(file):
    """Get timestamp from file."""
    timestamp = file.metadata.get("timestamp")
    if not timestamp:
        timestamp = file.source_created_time
    if not timestamp:
        # This definitely exists
        timestamp = file.uploaded_time
    return timestamp


def to_camel_case(snake):
    """Convert from snake case to camel case."""
    components = snake.split("_")
    return components[0] + "".join(x.title() for x in components[1:]) if len(components) > 1 else components[0]


def handle(data, client):
    """Gauge reader handle. Only analog, digital and level gauges supported at the moment."""
    if not {"gauge_type", "input_label", "output_label", "success_label", "failed_label"} <= data.keys():
        raise RuntimeError(
            "Data should contain all keys:  'gauge_type', 'input_label', 'output_label', 'success_label', 'failed_label'"
        )

    if data["gauge_type"] not in SUPPORTED_GAUGES:
        raise NotImplementedError(f"Only {SUPPORTED_GAUGES} gauge reading supported, not {data['gauge_type']}")

    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=logging.BASIC_FORMAT)

    files = client.files.list(
        labels=LabelFilter(contains_all=[data["input_label"]]),
        limit=15,
        uploaded=True,
    )
    logger.info(f"Number of files to process with label {data['input_label']}: {len(files)}")

    create_missing_labels(client, [data["output_label"], data["success_label"], data["failed_label"]])
    upload_queue = TimeSeriesUploadQueue(
        client,
        max_upload_interval=1,
        create_missing=True,
        max_queue_size=100,
    )
    project = client.config.project
    for file in files:
        ts_metadata = False
        logger.info(f"Processing file with external id {file.external_id}, id {file.id}")
        file.metadata = {} if file.metadata is None else file.metadata

        # If the file has a timeseries external id, try to get a complete metadata set from the timeseries.
        # If the timeseries does not have all metadata, try to get a complete metadata set from the file.
        if "ts_external_id" in file.metadata:
            ts = client.time_series.retrieve(external_id=file.metadata.get("ts_external_id"))
            logger.info(f"Retrieve timeseries {file.metadata.get('ts_external_id')}")
            if not ts:
                handle_failed_upload(
                    client=client,
                    id=file.id,
                    error_message=f"Time series with metadata {file.metadata.get('ts_external_id')} does not exist",
                    data=data,
                )
                continue
            if ts.metadata and all(key in ts.metadata for key in METADATA[data["gauge_type"]]["all"]):
                metadata = to_input_metadata(METADATA[data["gauge_type"]]["all"], ts.metadata)
                ts_metadata = True
                logger.info(f"Metadata from timeseries: {metadata}")
                for key in METADATA[data["gauge_type"]]["all"]:
                    file.metadata[key] = str(ts.metadata[key])
                    if "unit" in ts.metadata:
                        file.metadata["unit"] = ts.metadata["unit"]

            elif all(key in file.metadata for key in METADATA[data["gauge_type"]]["all"]):
                metadata = to_input_metadata(METADATA[data["gauge_type"]]["all"], file.metadata)
                logger.info(f"Metadata from file: {metadata}")
                update_ts_metadata(client, file.metadata, METADATA[data["gauge_type"]]["all"])
                logger.info(f"Complete metadata from image: {metadata}")
                ts_metadata = True

        if not ts_metadata:
            if not all(key in file.metadata for key in METADATA[data["gauge_type"]]["required"]):
                handle_failed_upload(
                    client=client,
                    id=file.id,
                    error_message=f"Some required metadata fields missing. Required metadata is: {METADATA[data['gauge_type']]['required']}. Cannot process image.",
                    data=data,
                )
                continue

            # If we did not find complete metadata from time series or file, use (possibly incomplete) metadata from file.
            metadata = {}
            for key in METADATA[data["gauge_type"]]["all"]:
                if key in file.metadata:
                    metadata[to_camel_case(key)] = file.metadata[key]
            logger.info(f"Not complete metadata from image: {metadata}")

        # Check gauge_type and make api-call based on the gauge_type
        if data["gauge_type"] == "valve":
            res = call_vision_api_endpoint(
                client=client,
                url=f"/api/v1/projects/{project}/context/vision/extract",
                input_data={"items": [{"fileId": file.id}], "features": ["ValveDetection"]},
                max_get_job_attempts=MAX_GET_JOB_ATTEMPTS,
                headers={"cdf-version": "beta"},
            )
        else:
            res = call_vision_api_endpoint(
                client=client,
                url=f"/api/v1/projects/{project}/context/vision/extract",
                input_data={
                    "items": [{"fileId": file.id}],
                    "features": [GAUGE_FEATURE_TYPES[data["gauge_type"]]],
                    "parameters": {GAUGE_PARAMETERS_TYPES[data["gauge_type"]]: metadata},
                },
                max_get_job_attempts=MAX_GET_JOB_ATTEMPTS,
                headers={"cdf-version": "beta"},
            )

        print(f"RES: {res}")
        if not res:
            handle_failed_upload(
                client=client, id=file.id, error_message="API call failed. Did not get a response", data=data
            )
            continue

        gauge_reading_attributes = gauge_reading_attributes_from_response(res, data["gauge_type"])
        logger.info("-----------------------------GAUGE READING-------------------------------")
        logger.info(f"External id: {file.external_id}, Reading: {gauge_reading_attributes}")
        logger.info("-------------------------------------------------------------------------")
        if gauge_reading_attributes is None:
            handle_failed_upload(
                client=client,
                id=file.id,
                error_message=f"Failed to read gauge {file.external_id}. No gauge found in image.",
                data=data,
            )
            continue
        read_value = (
            gauge_reading_attributes["attributes"]["valveState"]["value"]
            if data["gauge_type"] == "valve"
            else gauge_reading_attributes[data["gauge_type"] + "GaugeValue"]["value"]
        )
        if read_value is None:
            handle_failed_upload(
                client=client,
                id=file.id,
                error_message=f"Failed to read gauge {file.external_id}. Could not read value.",
                data=data,
                metadata=gauge_reading_attributes.get("metadata", None),
            )
            continue

        # If the reading was successful
        if data["gauge_type"] == "valve":
            file.metadata["state"] = read_value
            logger.info(f"Predicted state: {file.metadata['state'] }")

        else:
            # If the readings were successful, draw the annotations on the image
            create_annotations(
                client=client,
                gauge_reading_result=res,
                file_id=file.id,
                gauge_type=data["gauge_type"],
                bounding_box_label=BOUNDING_BOX_LABELS[data["gauge_type"]],
                keypoint_label=KEYPOINT_LABELS[data["gauge_type"]],
            )

            if data["gauge_type"] == "digital":
                file.metadata["value"] = read_value
            else:
                file.metadata["value"] = f"{read_value:.2f}"

            for key in METADATA[data["gauge_type"]]["all"]:
                key_attribute = to_camel_case(key)
                if gauge_reading_attributes is not None:
                    key_object = gauge_reading_attributes.get(key_attribute, None)
                    if key_object is not None:
                        key_val = key_object.get("value", None)
                        if key_val is not None:
                            file.metadata[key] = str(key_val)
                        elif key_object is None and key == "dead_angle":
                            file.metadata[key] = "90"  # Dead angle is not returned when it is 90 (default)
            logger.info(f"Predicted value: {file.metadata['value']}")

        if "ts_external_id" in file.metadata and not ts_metadata:
            # If the reading was successful and timeseries metadata does not exist, add ts metadata from the reading.
            metadata = to_input_metadata(METADATA[data["gauge_type"]]["all"], file.metadata)
            update_ts_metadata(client, file.metadata, METADATA[data["gauge_type"]]["all"])

        if "ts_external_id" in file.metadata:
            # Upload datapoint to timeseries if the file has ts_external_id.
            timestamp = int(get_timestamp(file))
            if data["gauge_type"] == "valve":
                timeseries_value = 1.0 if read_value == "on" else 0.0 if read_value == "off" else -1.0
            else:
                timeseries_value = float(read_value)

            upload_queue.add_to_upload_queue(
                datapoints=[(timestamp, timeseries_value)],
                external_id=file.metadata.get("ts_external_id"),
            )

            try:
                position = _get_position_from_metadata_to_vec3f(file.metadata)
                if data["gauge_type"] in ["dial", "digital", "level"]:
                    # create a observation if value is above or below threshold
                    _value_threshold_max = float(ts.metadata.get("observation_threshold_max", "inf"))
                    _value_threshold_min = float(ts.metadata.get("observation_threshold_min", "-inf"))

                    if timeseries_value < _value_threshold_min:
                        create_apm_observation_from_reading(
                            client=client,
                            file_external_id=file.external_id,
                            timeseries_external_id=file.metadata.get("ts_external_id"),
                            value=timeseries_value,
                            timestamp=timestamp,
                            pose=position,
                            message="WARNING: Value below threshold",
                            apm_checklistitem_external_id=file.metadata.get("action_run_id"),
                        )
                    elif timeseries_value > _value_threshold_max:
                        create_apm_observation_from_reading(
                            client=client,
                            file_external_id=file.external_id,
                            timeseries_external_id=file.metadata.get("ts_external_id"),
                            value=timeseries_value,
                            timestamp=timestamp,
                            pose=position,
                            message="WARNING: Value over threshold",
                            apm_checklistitem_external_id=file.metadata.get("action_run_id"),
                        )
                elif data["gauge_type"] == "valve":
                    _expected_state = ts.metadata.get("expected_valve_state", "no_state_specified")
                    if (_expected_state == "on" or _expected_state == "1.0") and timeseries_value == 0.0:
                        create_apm_observation_from_reading(
                            client=client,
                            file_external_id=file.external_id,
                            timeseries_external_id=file.metadata.get("ts_external_id"),
                            value=timeseries_value,
                            timestamp=timestamp,
                            pose=position,
                            message="WARNING: Valve is off",
                            apm_checklistitem_external_id=file.metadata.get("action_run_id"),
                        )
                    elif (_expected_state == "off" or _expected_state == "0.0") and timeseries_value == 1.0:
                        create_apm_observation_from_reading(
                            client=client,
                            file_external_id=file.external_id,
                            timeseries_external_id=file.metadata.get("ts_external_id"),
                            value=timeseries_value,
                            timestamp=timestamp,
                            pose=position,
                            message="WARNING: Valve is on",
                            apm_checklistitem_external_id=file.metadata.get("action_run_id"),
                        )

            except Exception as e:
                logger.error(f"Failed to create APM observation: {e}")

        client.files.update(
            FileMetadataUpdate(id=file.id)
            .labels.remove(data["input_label"])
            .labels.add([data["output_label"], data["success_label"]])
            .metadata.add(file.metadata)
        )
        logger.info(f"Gauge reading completed successfully for file {file.external_id}, id: {file.id}.")

    upload_queue.upload()
    return {}



import pytest
from cognite.client.data_classes.raw import RowWrite, RowWriteList

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import (
    BooleanProfileColumn,
    NumberProfileColumn,
    ObjectProfileColumn,
    StringProfileColumn,
    UnknownTypeProfileColumn,
    VectorProfileColumn,
)


@pytest.fixture()
def raw_data() -> RowWriteList:
    return RowWriteList(
        [
            RowWrite(
                key=f"row{i}",
                columns={
                    "StringCol": f"value{i % 3}",
                    "IntegerCol": i % 5,
                    "BooleanCol": [True, False][i % 2],
                    "FloatCol": i * 0.1,
                    "EmptyCol": None,
                    "ArrayCol": [i, i + 1, i + 2] if i % 2 == 0 else None,
                    "ObjectCol": {"nested_key": f"nested_value_{i}" if i % 2 == 0 else None},
                },
            )
            for i in range(10)
        ]
    )


class TestRawProfile:
    def test_raw_profile(self, toolkit_client: ToolkitClient, raw_data: RowWriteList):
        db_name = "toolkit_test_db"
        table_name = "toolkit_test_profiling_table"
        existing_dbs = toolkit_client.raw.databases.list(limit=-1)
        existing_table_names: set[str] = set()
        if db_name in {db.name for db in existing_dbs}:
            existing_table_names = {table.name for table in toolkit_client.raw.tables.list(db_name=db_name, limit=-1)}
        if table_name not in existing_table_names:
            toolkit_client.raw.rows.insert(db_name, table_name, raw_data, ensure_parent=True)

        limit = len(raw_data) // 2
        results = toolkit_client.raw.profile(db_name, table_name, limit=limit)
        assert results.row_count == limit
        assert results.column_count == len(raw_data[0].columns)
        assert isinstance(results.columns["StringCol"], StringProfileColumn)
        assert isinstance(results.columns["IntegerCol"], NumberProfileColumn)
        assert isinstance(results.columns["BooleanCol"], BooleanProfileColumn)
        assert isinstance(results.columns["FloatCol"], NumberProfileColumn)
        assert isinstance(results.columns["EmptyCol"], UnknownTypeProfileColumn)
        assert isinstance(results.columns["ArrayCol"], VectorProfileColumn)
        assert isinstance(results.columns["ObjectCol"], ObjectProfileColumn)



from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import data_modeling as dm

from cognite_toolkit._cdf_tk.client.data_classes.graphql_data_models import GraphQLDataModel, GraphQLDataModelWriteList
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.exceptions import ToolkitCycleError
from cognite_toolkit._cdf_tk.loaders import DataModelLoader, ResourceWorker
from cognite_toolkit._cdf_tk.loaders._resource_loaders import GraphQLLoader, ViewLoader
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestDataModelLoader:
    def test_update_data_model_random_view_order(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient
    ):
        cdf_data_model = dm.DataModel(
            space="sp_space",
            external_id="my_model",
            version="1",
            views=[
                dm.ViewId(space="sp_space", external_id="first", version="1"),
                dm.ViewId(space="sp_space", external_id="second", version="1"),
            ],
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            is_global=False,
        )
        # Simulating that the data model is available in CDF
        toolkit_client_approval.append(dm.DataModel, cdf_data_model)

        local_data_model = dm.DataModelApply(
            space="sp_space",
            external_id="my_model",
            version="1",
            views=[
                dm.ViewId(space="sp_space", external_id="second", version="1"),
                dm.ViewId(space="sp_space", external_id="first", version="1"),
            ],
            description=None,
            name=None,
        ).dump_yaml()

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_data_model

        loader = DataModelLoader.create_loader(
            env_vars_with_client.get_client(),
        )
        worker = ResourceWorker(loader)
        to_create, to_change, to_delete, unchanged = worker.prepare_resources([filepath])

        assert {
            "create": len(to_create),
            "change": len(to_change),
            "delete": len(to_delete),
            "unchanged": len(unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    def test_are_equal_version_int(self, env_vars_with_client: EnvironmentVariables) -> None:
        local_yaml = """space: sp_space
externalId: my_model
version: 1
views:
  - space: sp_space
    externalId: first
    version: 1
    type: view
        """
        cdf_data_model = dm.DataModel(
            space="sp_space",
            external_id="my_model",
            version="1",
            views=[dm.ViewId(space="sp_space", external_id="first", version="1")],
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            is_global=False,
        )
        loader = DataModelLoader.create_loader(env_vars_with_client.get_client())
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_yaml
        # The load filepath method ensures version is read as an int.
        local_dict = loader.load_resource_file(filepath, {})[0]

        cdf_dumped = loader.dump_resource(cdf_data_model, local_dict)

        assert local_dict == cdf_dumped


class TestGraphQLLoader:
    def test_deployment_order(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient
    ) -> None:
        loader = GraphQLLoader.create_loader(env_vars_with_client.get_client())
        # The first model is dependent on the second model
        first_file = self._create_mock_file(
            """
type WindTurbine @import @view(space: "second_space", externalId: "GeneratingUnit", version: "v1"){
name: String}""",
            "first_space",
            "WindTurbineModel",
        )
        second_file = self._create_mock_file(
            """
type GeneratingUnit {
        name: String
            }""",
            "second_space",
            "GeneratingUnitModel",
        )

        items = loader.load_resource_file(first_file, {})
        items.extend(loader.load_resource_file(second_file, {}))

        resources = GraphQLDataModelWriteList([loader.load_resource(item) for item in items])

        loader.create(resources)

        created = toolkit_client_approval.created_resources_of_type(GraphQLDataModel)

        assert len(created) == 2
        assert created[0].external_id == "GeneratingUnitModel"
        assert created[1].external_id == "WindTurbineModel"

    def test_raise_cycle_error(
        self, env_vars_with_client: EnvironmentVariables, toolkit_client_approval: ApprovalToolkitClient
    ) -> None:
        loader = GraphQLLoader.create_loader(env_vars_with_client.get_client())
        # The two models are dependent on each other
        first_file = self._create_mock_file(
            """type WindTurbine @import(dataModel: {externalId: "SolarModel", version: "v1", space: "second_space"}) {
name: String}""",
            "first_space",
            "WindTurbineModel",
        )
        second_file = self._create_mock_file(
            """type Solar @import(dataModel: {externalId: "WindTurbineModel", version: "v1", space: "first_space"}) {
        name: String
            }""",
            "second_space",
            "SolarModel",
        )

        items = loader.load_resource_file(first_file, {})
        items.extend(loader.load_resource_file(second_file, {}))
        resources = GraphQLDataModelWriteList([loader.load_resource(item) for item in items])
        with pytest.raises(ToolkitCycleError) as e:
            loader.create(resources)

        assert "Cycle detected" in str(e.value)
        assert [m.external_id for m in e.value.args[1]] == [
            "WindTurbineModel",
            "SolarModel",
            "WindTurbineModel",
        ]

    def test_load_version_int(self, env_vars_with_client: EnvironmentVariables) -> None:
        file = self._create_mock_file(
            """type WindTurbine{
            name: String}""",
            "DG-COR-ALL-DMD",
            "AssetHierarchyDOM",
            "3_0_2",
        )
        loader = GraphQLLoader.create_loader(env_vars_with_client.get_client())

        items = loader.load_resource_file(file, {})

        assert len(items) == 1
        resource = loader.load_resource(items[0], is_dry_run=False)
        assert resource.version == "3_0_2"

    @staticmethod
    def _create_mock_file(model: str, space: str, external_id: str, version: int | str = "v1") -> MagicMock:
        yaml_file = MagicMock(spec=Path)
        yaml_file.read_text.return_value = f"""space: {space}
externalId: {external_id}
version: {version}
dml: model.graphql
"""

        graphql_file = MagicMock(spec=Path)
        graphql_file.read_text.return_value = model
        graphql_file.name = "model.graphql"
        graphql_file.is_file.return_value = True

        yaml_file.with_suffix.return_value = graphql_file
        return yaml_file


@pytest.fixture()
def parent_grandparent_view() -> dm.ViewlList:
    return dm.ViewList(
        [
            dm.View(
                space="space",
                external_id="Parent",
                version="v1",
                name="Parent",
                description=None,
                implements=[dm.ViewId("space", "GrandParent", "v1")],
                properties={},
                last_updated_time=1,
                created_time=1,
                filter=None,
                writable=True,
                used_for="node",
                is_global=False,
            ),
            dm.View(
                space="space",
                external_id="GrandParent",
                version="v1",
                name="GrandParent",
                description=None,
                implements=[],
                properties={},
                last_updated_time=1,
                created_time=1,
                filter=None,
                writable=True,
                used_for="node",
                is_global=False,
            ),
        ]
    )


class TestViewLoader:
    def test_topological_sorting(self, parent_grandparent_view: dm.ViewList) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.views.retrieve.return_value = parent_grandparent_view
            loader = ViewLoader(client, Path("build_dir"), None, topological_sort_implements=True)
            actual = loader.topological_sort(
                [dm.ViewId("space", "Parent", "v1"), dm.ViewId("space", "GrandParent", "v1")]
            )

        assert actual == [dm.ViewId("space", "GrandParent", "v1"), dm.ViewId("space", "Parent", "v1")]

    def test_topological_sorting_cycle(self, parent_grandparent_view: dm.ViewList) -> None:
        parent_grandparent_view[1].implements = [parent_grandparent_view[0].as_id()]

        with monkeypatch_toolkit_client() as client, pytest.raises(ToolkitCycleError) as exc_info:
            client.data_modeling.views.retrieve.return_value = parent_grandparent_view
            loader = ViewLoader(client, Path("build_dir"), None, topological_sort_implements=True)
            loader.topological_sort([dm.ViewId("space", "Parent", "v1"), dm.ViewId("space", "GrandParent", "v1")])

        assert "cycle in implements" in str(exc_info.value)



import pytest
from cognite.client import data_modeling as dm

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.loaders import ContainerLoader, SpaceLoader, ViewLoader
from tests.data import STRONGLY_COUPLED_MODEL


@pytest.fixture()
def deployed_container_space_coupled_model(toolkit_client: ToolkitClient) -> None:
    space_loader = SpaceLoader(toolkit_client, STRONGLY_COUPLED_MODEL, None)
    files = space_loader.find_files()
    assert len(files) == 1
    space = dm.SpaceApply.load(files[0].read_text(encoding="utf-8"))
    if not space_loader.retrieve([space.as_id()]):
        created_space = space_loader.create([space])
        assert len(created_space) == 1

    container_loader = ContainerLoader(toolkit_client, STRONGLY_COUPLED_MODEL, None)
    containers = dm.ContainerApplyList(
        [dm.ContainerApply.load(file.read_text(encoding="utf-8")) for file in container_loader.find_files()]
    )

    if len(container_loader.retrieve(containers.as_ids())) != len(containers):
        created_containers = container_loader.create(containers)
        assert len(created_containers) == len(containers)


@pytest.mark.usefixtures("deployed_container_space_coupled_model")
def test_deploy_strongly_coupled_model(toolkit_client: ToolkitClient) -> None:
    loader = ViewLoader(toolkit_client, STRONGLY_COUPLED_MODEL, None)
    views = dm.ViewApplyList([dm.ViewApply.load(file.read_text(encoding="utf-8")) for file in loader.find_files()])

    try:
        created = loader.create(views)

        assert len(views) == len(created)
    finally:
        toolkit_client.data_modeling.views.delete(views.as_ids())



from pathlib import Path
from typing import Any

import pyperclip
import pytest
from cognite.client._api.iam import IAMAPI
from cognite.client.data_classes import GroupWrite, capabilities

from cognite_toolkit._cdf_tk.commands import AuthCommand, BuildCommand
from cognite_toolkit._cdf_tk.data_classes import (
    BuildConfigYAML,
    BuiltModule,
)
from cognite_toolkit._cdf_tk.loaders import GroupAllScopedLoader
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump

SKIP_ACLS = frozenset(
    {
        capabilities.LegacyGenericsAcl,
        capabilities.LegacyModelHostingAcl,
        capabilities.ExperimentsAcl,
        capabilities.ProjectsAcl,
        capabilities.UserProfilesAcl,
    }
)

READ_ACTIONS = frozenset(
    {
        "READ",
        "USE",
        "LIST",
        "MEMBEROF",
    }
)


@pytest.fixture(scope="session")
def built_module(tmp_path_factory: Any, organization_dir: Path) -> BuiltModule:
    cmd = BuildCommand(silent=True, skip_tracking=True)
    config = BuildConfigYAML.load_default(organization_dir)
    config.environment.selected = ["cdf_auth_readwrite_all"]
    built = cmd.build_config(
        build_dir=tmp_path_factory.mktemp("build"),
        organization_dir=organization_dir,
        config=config,
        packages={},
        on_error="raise",
    )
    assert len(built) == 1

    return built[0]


@pytest.fixture
def read_write_group(built_module: BuiltModule, env_vars_with_client: EnvironmentVariables) -> GroupWrite:
    auth = built_module.resources["auth"]
    read_write_resource = next((resource for resource in auth if resource.identifier == "gp_admin_read_write"), None)
    assert read_write_resource is not None

    loader = GroupAllScopedLoader(env_vars_with_client.get_client(), None, None)
    items = loader.load_resource_file(read_write_resource.destination, env_vars_with_client.dump())
    return loader.load_resource(items[0], is_dry_run=True)


@pytest.fixture
def readonly_group(built_module: BuiltModule, env_vars_with_client: EnvironmentVariables) -> GroupWrite:
    auth = built_module.resources["auth"]
    readonly_resource = next((resource for resource in auth if resource.identifier == "gp_admin_readonly"), None)
    assert readonly_resource is not None

    loader = GroupAllScopedLoader(env_vars_with_client.get_client(), None, None)
    items = loader.load_resource_file(readonly_resource.destination, env_vars_with_client.dump())
    return loader.load_resource(items[0], is_dry_run=True)


def get_all_capabilities(readonly: bool = False) -> list[capabilities.Capability]:
    all_capabilities: list[capabilities.Capability] = []
    for name, capability_cls in capabilities._CAPABILITY_CLASS_BY_NAME.items():
        if capability_cls in SKIP_ACLS:
            continue
        actions = list(capability_cls.Action)
        if readonly:
            actions = [action for action in actions if action.value in READ_ACTIONS]
        available_scopes = vars(capability_cls.Scope)
        if "All" not in available_scopes:
            raise ValueError(f"Capability {capability_cls} does not have a scope named 'All'")
        scope = available_scopes["All"]

        capability = capability_cls(actions=actions, scope=scope())
        all_capabilities.append(capability)
    return all_capabilities


class TestCDFAuthReadWriteAll:
    def test_read_write_group_is_up_to_date(self, read_write_group: GroupWrite) -> None:
        missing_capabilities = IAMAPI.compare_capabilities(
            read_write_group.capabilities, get_all_capabilities(readonly=False)
        )
        if missing_capabilities:
            merged = AuthCommand._merge_capabilities(missing_capabilities)
            missing_yaml = yaml_safe_dump([item.dump() for item in merged], indent=2)
            pyperclip.copy(missing_yaml)

        # This test will fail typically after you have updated the cognite-sdk-python package.
        # It checks that all capabilities that are available in the SDK are also available in the group
        # 'gp_admin_read_write'.
        # For your convenience, the missing capabilities will be copied to your clipboard. You
        # need to go to the 'cognite_toolkit/builtin_modules' and find the 'cdf_auth_readwrite_all' module,
        # and update the capabilities in the 'gp_admin_read_write' group.
        assert not missing_capabilities, (
            f"Missing {len(missing_capabilities)} the missing capabilities have been copied to your clipboard."
        )

    def test_readonly_group_is_up_to_date(self, readonly_group: GroupWrite) -> None:
        missing_capabilities = IAMAPI.compare_capabilities(
            readonly_group.capabilities, get_all_capabilities(readonly=True)
        )
        if missing_capabilities:
            merged = AuthCommand._merge_capabilities(missing_capabilities)
            missing_yaml = yaml_safe_dump([item.dump() for item in merged], indent=2)
            pyperclip.copy(missing_yaml)

        # Similar to the previous test, this test will fail typically after you have updated the
        # cognite-sdk-python package.
        # Running this you need to update the 'gp_admin_readonly' group in the 'cdf_auth_readwrite_all' module.
        assert not missing_capabilities, (
            f"Missing {len(missing_capabilities)} the missing capabilities have been copied to your clipboard"
        )



from __future__ import annotations

import dataclasses
import itertools
import warnings
from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Collection, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from functools import total_ordering
from typing import Any, ClassVar, Generic, TypeVar

from rich import print
from rich.console import Console

RICH_WARNING_FORMAT = "    [bold yellow]WARNING:[/] "
RICH_WARNING_DETAIL_FORMAT = f"{'    ' * 2}"


class SeverityLevel(Enum):
    ERROR = "error"
    HIGH = "red"
    MEDIUM = "yellow"
    LOW = "green"

    @property
    def prefix(self) -> str:
        if self == SeverityLevel.ERROR:
            return f"[bold red]ERROR [{self.name}]:[/]"
        else:
            return f"[bold {self.value}]WARNING [{self.name}]:[/]"

    @property
    def prefix_length(self) -> int:
        return len(self.prefix.split("]", 1)[1].rsplit("[", 1)[0])


@total_ordering
@dataclass(frozen=True)
class ToolkitWarning(ABC, UserWarning):
    severity: ClassVar[SeverityLevel]

    def group_key(self) -> tuple[Any, ...]:
        """This is used to group warnings together when printing them out."""
        return (type(self).__name__,)

    def group_header(self) -> str:
        """This can be overridden to provide a custom header for a group of warnings."""
        return f"    {type(self).__name__}:"

    def as_tuple(self) -> tuple[Any, ...]:
        return type(self).__name__, *dataclasses.astuple(self)

    def __lt__(self, other: ToolkitWarning) -> bool:
        if not isinstance(other, ToolkitWarning):
            return NotImplemented
        return self.as_tuple() < other.as_tuple()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ToolkitWarning):
            return NotImplemented
        return self.as_tuple() == other.as_tuple()

    @abstractmethod
    def get_message(self) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.get_message()

    def print_prepare(self) -> tuple[str, str]:
        prefix = self.severity.prefix
        end = "\n" + " " * ((self.severity.prefix_length + 1) // 2)
        message = self.get_message().replace("\n", end)
        return prefix, message

    def print_warning(self, include_timestamp: bool = False, console: Console | None = None) -> None:
        parts: tuple[str, ...] = self.print_prepare()
        if include_timestamp:
            timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S+00:00")
            parts = (timestamp, *parts)
        if console is None:
            print(*parts)
        else:
            console.print(*parts)


T_Warning = TypeVar("T_Warning", bound=ToolkitWarning)


class WarningList(UserList, Generic[T_Warning]):
    def __init__(self, collection: Collection[T_Warning] | None = None) -> None:
        super().__init__(collection or [])

    def __str__(self) -> str:
        output = [""]
        for group_key, group in itertools.groupby(sorted(self), key=lambda w: w.group_key()):
            group_list = list(group)
            header = group_list[0].group_header()
            if header:
                output.append(header)
            space = "    " * 2
            for warning in group_list:
                warning_str = str(warning).replace("\n", f"\n{space}")
                output.append(f"{space} * {warning_str}")
        return "\n".join(output)


@dataclass(frozen=True)
class GeneralWarning(ToolkitWarning, ABC):
    severity: ClassVar[SeverityLevel]
    message: ClassVar[str | None] = None


@contextmanager
def catch_warnings(warning_type: type[ToolkitWarning] | None = None) -> Iterator[WarningList[ToolkitWarning]]:
    """Catch warnings and append them to the warning list."""
    warning_list = WarningList[ToolkitWarning]()
    with warnings.catch_warnings(record=True) as warning_logger:
        try:
            yield warning_list
        finally:
            for warning in warning_logger:
                if warning_type is None or isinstance(warning.message, warning_type):
                    warning_list.append(warning.message)
                elif isinstance(warning.message, ToolkitWarning):
                    warning.message.print_warning()
                else:
                    warnings.warn(warning.message, stacklevel=2)



from __future__ import annotations

import re
import warnings
from collections.abc import Hashable
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic

from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
)
from rich import print
from rich.panel import Panel
from yaml import YAMLError

from cognite_toolkit._cdf_tk.constants import TABLE_FORMATS
from cognite_toolkit._cdf_tk.exceptions import ToolkitWrongResourceError, ToolkitYAMLFormatError
from cognite_toolkit._cdf_tk.tk_warnings import EnvironmentVariableMissingWarning, catch_warnings
from cognite_toolkit._cdf_tk.utils import to_diff

from ._base_loaders import T_ID, ResourceLoader, T_WritableCogniteResourceList

if TYPE_CHECKING:
    from cognite_toolkit._cdf_tk.data_classes._module_directories import ReadModule


class ResourceWorker(
    Generic[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList]
):
    def __init__(
        self,
        loader: ResourceLoader[
            T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList
        ],
    ):
        self.loader = loader
        self.duplicates: list[T_ID] = []

    def load_files(
        self, sort: bool = True, directory: Path | None = None, read_modules: list[ReadModule] | None = None
    ) -> list[Path]:
        filepaths = self.loader.find_files(directory)

        for read_module in read_modules or []:
            if resource_dir := read_module.resource_dir_path(self.loader.folder_name):
                # As of 05/11/24, Asset support csv and parquet files in addition to YAML.
                # These table formats are not built, i.e., no variable replacement is done,
                # so we load them directly from the source module.
                filepaths.extend(self.loader.find_files(resource_dir, include_formats=TABLE_FORMATS))

        if not sort:
            return filepaths

        def sort_key(p: Path) -> int:
            if result := re.findall(r"^(\d+)", p.stem):
                return int(result[0])
            else:
                return len(filepaths)

        # In the build step, the resource files are prefixed a number that controls the order in which
        # the resources are deployed. The custom 'sort_key' here is to get a sort on integer instead of a default string
        # sort.
        return sorted(filepaths, key=sort_key)

    def prepare_resources(
        self,
        filepaths: list[Path],
        environment_variables: dict[str, str | None] | None = None,
        is_dry_run: bool = False,
        force_update: bool = False,
        verbose: bool = False,
    ) -> tuple[T_CogniteResourceList, T_CogniteResourceList, list[T_ID], T_CogniteResourceList]:
        """Prepare resources for deployment by loading them from files, validating access, and categorizing them into create, update, delete, and unchanged lists.

        Args:
            filepaths: The list of file paths to load resources from.
            environment_variables: Environment variables to use for variable replacement in the resource files.
            is_dry_run: Whether to perform a dry run (no actual changes made).
            force_update: Whether to force update existing resources even if they are unchanged.
            verbose: Whether to print detailed information about the resources being processed.

        Returns:
            A tuple containing:
                - to_create: List of resources to create.
                - to_update: List of resources to update.
                - to_delete: List of resource IDs to delete.
                - unchanged: List of resources that are unchanged.

        """
        local_by_id = self.load_resources(filepaths, environment_variables, is_dry_run)

        self.validate_access(local_by_id, is_dry_run)

        # Lookup the existing resources in CDF
        cdf_resources: T_WritableCogniteResourceList
        cdf_resources = self.loader.retrieve(list(local_by_id.keys()))
        return self.categorize_resources(local_by_id, cdf_resources, force_update, verbose)

    def load_resources(
        self, filepaths: list[Path], environment_variables: dict[str, str | None] | None, is_dry_run: bool
    ) -> dict[T_ID, tuple[dict[str, Any], T_WriteClass]]:
        local_by_id: dict[T_ID, tuple[dict[str, Any], T_WriteClass]] = {}  # type: ignore[assignment]
        # Load all resources from files, get ids, and remove duplicates.
        environment_variables = environment_variables or {}
        for filepath in filepaths:
            with catch_warnings(EnvironmentVariableMissingWarning) as warning_list:
                try:
                    resource_list = self.loader.load_resource_file(filepath, environment_variables)
                except YAMLError as e:
                    raise ToolkitYAMLFormatError(f"YAML validation error for {filepath.name}: {e}")
            identifiers: list[Hashable] = []
            for resource_dict in resource_list:
                identifier = self.loader.get_id(resource_dict)
                identifiers.append(identifier)
                try:
                    # The load resource modifies the resource_dict, so we deepcopy it to avoid side effects.
                    loaded = self.loader.load_resource(deepcopy(resource_dict), is_dry_run)
                except ToolkitWrongResourceError:
                    # The ToolkitWrongResourceError is a special exception that as of 21/12/24 is used by
                    # the GroupAllScopedLoader and GroupResourceScopedLoader to signal that the resource
                    # should be handled by the other loader.
                    continue
                if identifier in local_by_id:
                    self.duplicates.append(identifier)
                else:
                    local_by_id[identifier] = resource_dict, loaded

            for warning in warning_list:
                if isinstance(warning, EnvironmentVariableMissingWarning):
                    # Warnings are immutable, so we use the below method to override it.
                    object.__setattr__(warning, "identifiers", frozenset(identifiers))
                    # Reraise the warning to be caught higher up.
                    warnings.warn(warning, stacklevel=2)
                else:
                    warning.print_warning()
        return local_by_id

    def validate_access(self, local_by_id: dict[T_ID, tuple[dict[str, Any], T_WriteClass]], is_dry_run: bool) -> None:
        capabilities = self.loader.get_required_capability(
            [item for _, item in local_by_id.values()], read_only=is_dry_run
        )
        if capabilities and (missing := self.loader.client.verify.authorization(capabilities)):
            raise self.loader.client.verify.create_error(missing, action=f"clean {self.loader.display_name}")

    def categorize_resources(
        self,
        local_by_id: dict[T_ID, tuple[dict[str, Any], T_WriteClass]],
        cdf_resources: T_WritableCogniteResourceList,
        force_update: bool,
        verbose: bool,
    ) -> tuple[T_CogniteResourceList, T_CogniteResourceList, list[T_ID], T_CogniteResourceList]:
        to_create: T_CogniteResourceList
        to_update: T_CogniteResourceList
        to_delete: list[T_ID] = []
        unchanged: T_CogniteResourceList
        to_create, to_update, unchanged = (
            self.loader.list_write_cls([]),
            self.loader.list_write_cls([]),
            self.loader.list_write_cls([]),
        )
        cdf_resource_by_id = {self.loader.get_id(resource): resource for resource in cdf_resources}
        for identifier, (local_dict, local_resource) in local_by_id.items():
            cdf_resource = cdf_resource_by_id.get(identifier)
            if cdf_resource is None:
                to_create.append(local_resource)
                continue
            cdf_dict = self.loader.dump_resource(cdf_resource, local_dict)
            if not force_update and cdf_dict == local_dict:
                unchanged.append(local_resource)
                continue
            if self.loader.support_update:
                to_update.append(local_resource)
            else:
                to_delete.append(identifier)
                to_create.append(local_resource)
            if verbose:
                diff_str = "\n".join(to_diff(cdf_dict, local_dict))
                for sensitive in self.loader.sensitive_strings(local_resource):
                    diff_str = diff_str.replace(sensitive, "********")
                print(
                    Panel(
                        diff_str,
                        title=f"{self.loader.display_name}: {identifier}",
                        expand=False,
                    )
                )
        return to_create, to_update, to_delete, unchanged



from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import MigrationPrepareCommand


class TestMigrateTimeSeriesCommand:
    def test_migration_prepare_command(
        self,
        toolkit_client: ToolkitClient,
    ) -> None:
        cmd = MigrationPrepareCommand(silent=True, skip_tracking=True)

        dry_run_result = cmd.deploy_cognite_migration(toolkit_client, True, verbose=False)

        actual_result = cmd.deploy_cognite_migration(toolkit_client, False, verbose=False)

        assert dry_run_result.has_counts, "Dry run should have counts"
        assert actual_result.has_counts, "Actual should have counts"



from pathlib import Path

raw = """name: encoding_confusion
query: >-
    SELECT * FROM my_éñcüd€d£d_table
    WHERE column = 'value'
destination:
  type: assets
ignoreNullFields: true
isPublic: true
conflictMode: upsert
dataSetExternalId: ds_at_p66
"""

THIS_FOLDER = Path(__file__).resolve(strict=True).parent
encodings = ["utf-8", "utf-16", "cp1252"]
for encoding in encodings:
    file = THIS_FOLDER / f"my_{encoding}.Transformation.yaml"
    file.write_text(f"externalId: encoding_transformation_{encoding}\n{raw}", encoding=encoding)



#!/usr/bin/env python

# You use this file to run the cdf-tk file for development purposes in Visual Studio Code
# to avoid installing the cognite-toolkit package.
# cdf.py is found inside the cognite-toolkit package, which is fine when you do pip install cognite-toolkit
# However, when you run the file in Visual Studio Code, the module should not be installed in your
# python virtual environment, but rather be found in the root of the repo.
# This workaround allows you to run cdf.py in Visual Studio Code like this:
"""{
    "name": "Python: build",
    "type": "python",
    "request": "launch",
    "program": "./cdf-tk-dev.py",
    "args": [
        "--verbose",
        "--override-env",
        "build",
        "--build-dir=build",
        "--clean",
        "--env=local",
        "./cognite_toolkit/"
    ],
    "console": "integratedTerminal",
    "justMyCode": false
},
"""

import os
import sys
from pathlib import Path

from cognite_toolkit._cdf_tk.data_classes import Environment, InitConfigYAML

REPO_ROOT = Path(__file__).resolve().parent
root_folder = rf"{REPO_ROOT.absolute()}"

sys.path.append(root_folder)
# Avoid sending requests to sentry when doing development
os.environ["SENTRY_ENABLED"] = "false"

from cognite_toolkit._cdf import app  # noqa: E402

if not Path(REPO_ROOT / "cognite_toolkit" / "config.local.yaml").exists():
    config_init = InitConfigYAML(
        Environment(
            name="local",
            project="<my-project>",
            validation_type="dev",
            selected=["cdf_demo_infield", "cdf_oid_example_data"],
        )
    ).load_defaults(REPO_ROOT / "cognite_toolkit")
    (REPO_ROOT / "cognite_toolkit" / "config.local.yaml").write_text(config_init.dump_yaml_with_comments())

if __name__ == "__main__":
    app()



import sys
from typing import cast

from cognite.client._http_client import _RetryTracker

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ToolkitRetryTracker(_RetryTracker):
    def copy(self) -> Self:
        new_tracker = ToolkitRetryTracker(self.config)
        new_tracker.read = self.read
        new_tracker.connect = self.connect
        new_tracker.status = self.status
        return cast(Self, new_tracker)



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import TransformationScheduleYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestTransformationScheduleYAML:
    @pytest.mark.parametrize("data", list(find_resources("schedule", "transformations")))
    def test_load_valid_transformation_schedule(self, data: dict[str, object]) -> None:
        loaded = TransformationScheduleYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data, expected_errors",
        [
            pytest.param(
                {"externalId": "fn_first_function"},
                ["Missing required field: 'interval'"],
                id="missing_required_field",
            )
        ],
    )
    def test_load_invalid_transformation_schedule(self, data: dict[str, object], expected_errors: list[str]) -> None:
        warnings = validate_resource_yaml_pydantic(data, TransformationScheduleYAML, source_file=Path("test.yaml"))
        assert len(warnings) == 1, "Only one warning should be raised"
        warning = warnings[0]
        assert isinstance(warning, ResourceFormatWarning), (
            f"Warning should be of type ResourceFormatWarning but got {type(warning)}"
        )
        assert list(warning.errors) == expected_errors, (
            f"Warning should contain the expected error {expected_errors} but got {list(warning.errors)}"
        )



from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.commands._migrate.data_classes import IdMigrationMapping, MigrationMappingList
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence


class TestChunkSequence:
    def test_chunk_sequence(self):
        """Test that chunker_sequence correctly chunks a sequence."""
        sequence = list(range(10))
        chunk_size = 3
        expected_chunks = [
            [0, 1, 2],
            [3, 4, 5],
            [6, 7, 8],
            [9],
        ]

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert chunks == expected_chunks

    def test_chunk_sequence_empty(self):
        """Test that chunker_sequence returns an empty list for an empty sequence."""
        sequence = []
        chunk_size = 3

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert chunks == []

    def test_chunk_sequence_maintain_source_type(self) -> None:
        """Test that chunker_sequence maintains the type of the source sequence."""
        sequence = MigrationMappingList(
            [
                IdMigrationMapping(
                    resource_type="timeseries", id=1, data_set_id=123, instance_id=NodeId("sp_full_ts", "full_ts_id")
                ),
                IdMigrationMapping(
                    resource_type="timeseries", id=2, data_set_id=None, instance_id=NodeId("sp_step_ts", "step_ts_id")
                ),
            ]
        )
        chunk_size = 1

        chunks = list(chunker_sequence(sequence, chunk_size))
        assert len(chunks) == 2
        assert len(chunks[0]) == 1
        assert len(chunks[1]) == 1
        assert isinstance(chunks[0], MigrationMappingList)



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.prototypes.commands.import_ import ImportTransformationCLI
from tests.data import TRANSFORMATION_CLI


class TestImportTransformationCLI:
    @pytest.mark.parametrize("source_name", ["manifest", "manifest_to_asset"])
    def test_import_transformation_cli(self, source_name: str, tmp_path: Path) -> None:
        transformation, schedule, notification, sql = (
            f"{source_name}.Transformation",
            f"{source_name}.Schedule",
            f"{source_name}.Notification",
            f"{source_name}.Transformation.sql",
        )
        expected = {transformation, schedule, notification}
        cmd = ImportTransformationCLI(print_warning=False)
        cmd.execute(TRANSFORMATION_CLI / f"{source_name}.yaml", tmp_path, False, False, False, verbose=False)

        files_by_name = {file.stem: file for file in tmp_path.rglob("*")}
        assert len(files_by_name) == len(expected)
        missing = expected - set(files_by_name)
        assert not missing, f"Missing files: {missing}"
        for name in expected:
            assert (tmp_path / f"{name}.yaml").read_text() == (TRANSFORMATION_CLI / f"{name}.yaml").read_text()
        if (tmp_path / sql).exists():
            assert (tmp_path / sql).read_text() == (TRANSFORMATION_CLI / sql).read_text()



import re
from collections.abc import Iterator
from dataclasses import dataclass
from typing import TypeAlias

from cognite_toolkit._cdf_tk.utils.file import YAMLComment

yaml_key: TypeAlias = tuple[str | int, ...]


class YAMLComments:
    def __init__(self, comments: dict[yaml_key, YAMLComment]) -> None:
        self.comments = comments

    @classmethod
    def load(cls, yaml_str: str) -> "YAMLComments":
        return cls(_YAMLCommentParser(yaml_str).parse())

    def dump(self, yaml_str: str) -> str:
        return _YAMLCommentParser(yaml_str).dump(self.comments)


@dataclass
class _YAMLLine:
    indent: int
    key: str | None
    comment: str | None
    is_array: bool
    raw: str

    @property
    def is_comment_line(self) -> bool:
        return self.comment is not None and self.key is None


class _YAMLCommentParser:
    token_pattern = re.compile(r"[\w\n]+|[^\w\s]", flags=re.DOTALL)

    def __init__(self, yaml_str: str) -> None:
        self.yaml_str = yaml_str

    def _iterate_lines(self) -> Iterator[tuple[list[str | int], _YAMLLine]]:
        lines = self.yaml_str.splitlines()
        key: list[str | int] = []
        last_indent = 0
        indent_size: int | None = None
        index_key_level_by_indent: dict[int, tuple[int, int]] = {}
        for line_str in lines:
            line = self._parse_line(line_str)
            if indent_size is None and line.indent > 0:
                # Infer indentation size
                indent_size = line.indent

            if line.is_comment_line:
                ...
            elif line.is_array:
                if line.indent not in index_key_level_by_indent:
                    index, key_level = 0, len(key)
                else:
                    index, key_level = index_key_level_by_indent[line.indent]
                    index += 1
                index_key_level_by_indent[line.indent] = index, key_level
                key.append(index)
                if line.key:
                    key.append(line.key)
            elif line.key and line.indent > last_indent:
                key.append(line.key)
            elif line.key and line.indent < last_indent:
                if indent_size is None:
                    raise ValueError("Indentation size could not be determined")
                key = key[: line.indent // indent_size]
                key.append(line.key)
            elif line.key:
                if key:
                    key.pop()
                key.append(line.key)

            yield key, line
            if not line.is_comment_line:
                last_indent = line.indent

    def parse(self) -> dict[yaml_key, YAMLComment]:
        comments: dict[yaml_key, YAMLComment] = {}
        last_comment: list[str] = []
        for full_key, line in self._iterate_lines():
            if line.comment and line.key:
                # End-of-line comment
                comments[tuple(full_key)] = YAMLComment(after=[line.comment], above=last_comment)
                last_comment = []
            elif line.comment:
                last_comment.append(line.comment)
            elif line.key and last_comment:
                comments[tuple(full_key)] = YAMLComment(above=last_comment)
                last_comment = []

        return comments

    def dump(self, comments: dict[yaml_key, YAMLComment]) -> str:
        new_lines: list[str] = []
        for full_key, line in self._iterate_lines():
            if comment := comments.get(tuple(full_key)):
                for above_comment in comment.above:
                    new_lines.append(f"{' ' * line.indent}#{above_comment}")
                if comment.after:
                    after_comments = " ".join(comment.after)
                    new_lines.append(f"{line.raw} #{after_comments}")
                else:
                    new_lines.append(line.raw)
            else:
                new_lines.append(line.raw)
        return "\n".join(new_lines)

    def _parse_line(self, line_str: str) -> _YAMLLine:
        indent = len(line_str) - len(line_str.lstrip())
        is_array = False
        key: str | None = None
        comment: str | None = None
        in_single_quote = False
        in_double_quote = False
        for no, character in enumerate(line_str):
            if character == '"':
                in_double_quote = not in_double_quote
            elif character == "'":
                in_single_quote = not in_single_quote

            if in_single_quote or in_double_quote:
                continue

            if character == "#":
                comment = "".join(line_str[no + 1 :])
                break
            elif character == ":" and no > 0:
                key = "".join(line_str[:no]).lstrip()
            elif character == "-":
                is_array = True
                indent = len(line_str) - len(line_str.lstrip().lstrip("-").lstrip())

        return _YAMLLine(indent, key, comment, is_array, line_str)



from __future__ import annotations

import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest
from cognite.client.data_classes import CreatedSession
from pytest import MonkeyPatch

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import ModulesCommand, RepoCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.constants import REPO_ROOT
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import PrintCapture

THIS_FOLDER = Path(__file__).resolve().parent
TMP_FOLDER = THIS_FOLDER / "tmp"
TMP_FOLDER.mkdir(exist_ok=True)


@pytest.fixture
def toolkit_client_approval() -> Iterator[ApprovalToolkitClient]:
    with monkeypatch_toolkit_client() as toolkit_client:

        def create_session(*args: Any, **kwargs: Any) -> CreatedSession:
            return CreatedSession(
                id=42,
                status="READY",
                nonce="dummy-nonce",
                type="CLIENT_CREDENTIALS",
            )

        toolkit_client.iam.sessions.create = create_session
        approval_client = ApprovalToolkitClient(toolkit_client)
        yield approval_client


@pytest.fixture(scope="function")
def env_vars_with_client(toolkit_client_approval: ApprovalToolkitClient) -> EnvironmentVariables:
    env_vars = EnvironmentVariables(
        CDF_CLUSTER="bluefield",
        CDF_PROJECT="pytest-project",
        LOGIN_FLOW="client_credentials",
        PROVIDER="entra_id",
        IDP_CLIENT_ID="dummy-123",
        IDP_CLIENT_SECRET="dummy-secret",
        IDP_TENANT_ID="dummy-domain",
    )
    env_vars._client = toolkit_client_approval.mock_client
    return env_vars


@pytest.fixture(scope="session")
def build_tmp_path() -> Iterator[Path]:
    pidid = os.getpid()
    build_folder = TMP_FOLDER / f"build-{pidid}"
    if build_folder.exists():
        shutil.rmtree(build_folder, ignore_errors=True)
        build_folder.mkdir(exist_ok=True)
    yield build_folder
    shutil.rmtree(build_folder, ignore_errors=True)


@pytest.fixture(scope="session")
def local_tmp_repo_path() -> Iterator[Path]:
    pidid = os.getpid()
    repo_path = TMP_FOLDER / f"pytest-repo-{pidid}"
    repo_path.mkdir(exist_ok=True)
    RepoCommand(silent=True, skip_git_verify=True).init(repo_path, host="GitHub")
    yield repo_path
    shutil.rmtree(repo_path, ignore_errors=True)


@pytest.fixture(scope="session")
def organization_dir(
    local_tmp_repo_path: Path,
) -> Path:
    organization_folder = "pytest-org"
    organization_dir = local_tmp_repo_path / organization_folder
    ModulesCommand(silent=True).init(
        organization_dir,
        select_all=True,
        clean=True,
    )

    return organization_dir


@pytest.fixture
def organization_dir_mutable(
    local_tmp_repo_path: Path,
) -> Path:
    """This is used in tests were the source module files are modified. For example, cdf pull commands."""
    organization_dir = local_tmp_repo_path / "pytest-org-mutable"

    ModulesCommand(silent=True).init(
        organization_dir,
        select_all=True,
        clean=True,
    )

    return organization_dir


@pytest.fixture
def capture_print(monkeypatch: MonkeyPatch) -> PrintCapture:
    capture = PrintCapture()
    toolkit_path = REPO_ROOT / "cognite_toolkit"
    builtin_modules_path = toolkit_path / "_cdf_tk" / "_packages"
    monkeypatch.setattr("cognite_toolkit._cdf.print", capture)

    # Monkeypatch all print functions in the toolkit automatically
    for folder in ["_cdf_tk", "_api"]:
        for py_file in (toolkit_path / folder).rglob("*.py"):
            if py_file.is_relative_to(builtin_modules_path):
                # Don't monkeypatch the function code of the
                # builtin modules
                continue
            file_path = py_file.relative_to(toolkit_path)
            module_path = f"{'.'.join(['cognite_toolkit', *file_path.parts[:-1]])}.{file_path.stem}"
            try:
                monkeypatch.setattr(f"{module_path}.print", capture)
            except AttributeError:
                # The module does not have a print function
                continue
            except ImportError:
                raise

    return capture



from __future__ import annotations

from rich import print
from rich.table import Table

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.feature_flags import Flags


class FeatureFlagCommand(ToolkitCommand):
    @staticmethod
    def list() -> None:
        user_settings = CDFToml.load().alpha_flags
        table = Table(title="feature flags")
        table.add_column("Name", justify="left")
        table.add_column("Description", justify="left")
        table.add_column("Status", justify="left")

        for flag in Flags:
            is_enabled = user_settings.get(flag.name, False)
            if is_enabled or flag.value.get("visible", False):
                table.add_row(
                    flag.name,
                    str(flag.value.get("description", "")),
                    "enabled" if is_enabled else "disabled",
                    style="yellow" if is_enabled else "",
                )
        print(table)
